# Copyright (c) 2019 Eric Steinberger


import copy
import time

import numpy as np
from gym import spaces

from PokerRL.game.Poker import Poker
from PokerRL.game.PokerEnvStateDictEnums import EnvDictIdxs, PlayerDictIdxs
from PokerRL.game._.rl_env.base._Deck import DeckOfCards
from PokerRL.game._.rl_env.base._PokerPlayer import PokerPlayer


class PokerEnv:
    """
    This abstract class implements general functions and attributes of a Poker Environment. Subclasses can define
    exact rules and hyperparameters of the environment. This base-engine supports HeadsUp and N-player poker.
    It is sub-base-classed by a Limit- and a Discrete- wrapper, as this base-class would default to the No-Limit
    ruleset.

    Functionality & Dynamics:
        reward:
            if terminal state:
                                numpy.array(shape=n_seats) that contains ""stack_after_episode - starting_stack""
                                for every player. This reward is either scaled or unscaled, depending on the setting
                                in the args that were passed when creating an instance of a PokerEnv subclass.

            else:
                                numpy.zeros(shape=n_seats)

        obs:
            if terminal state:
                                numpy array of zeros. The shape is the same as in non terminal states

            else:
                                covering PUBLIC state of the env (stacks, pot, board, etc.) No hole cards (or any
                                private information) from any player are included! To get a player's private
                                information, checkout .get_hole_cards_of_player()


        actions:
            For DiscretePokerEnv subclasses:
                                    0 = FOLD, 1 = CHECK/CALL
                                    2 = RAISE_SIZE_1, ... N = RAISE_SIZE_N-1

            For direct subclasses (i.e. "approximately continuous bet size" poker games like No Limit Holdem:
                                    Tuple: (action, raise size), where action is 0 for FOLD, 1 for CHECK/CALL, 2 for
                                    BET/RAISE. the raise_size is always to be passed, but only matters if the action
                                    is 2 (i.e. BET/RAISE). It is expressed as the total number of chips the player
                                    wants to have bet after placing the bet. So if the current bet is 30, and the agent
                                    wants to bet 60 chips more, the tuple should be (2, 90).

        card representations:
            Single cards can be referenced in 2 forms: 1D and 2D.
            1D refers to a card mapped to a single unique integer.
            2D refers to an array/tuple/list that contains the rank at the 1st index and the suit at the 2nd.

            A poker hand or the public board can be seen as an array of these cards. board_2d, for example refers to a
            2D array [[card1_rank, card1_suit], [card2_rank, card2_suit], ...].

            Alternatively, a hand of multiple cards can be referenced to by its index in an array sorted by the 1D
            representation of the first card, then by the 2nd, and so on. This format is called ""range_idx"" in this
            framework.
    """

    # _____________________ Variables to be defined by the subclassing game according to its rules _____________________

    SMALL_BLIND = NotImplementedError
    BIG_BLIND = NotImplementedError
    ANTE = NotImplementedError
    SMALL_BET = NotImplementedError
    BIG_BET = NotImplementedError
    DEFAULT_STACK_SIZE = NotImplementedError

    EV_NORMALIZER = NotImplementedError  # Divisor for chip numbers to compute e.g. MBB/H from raw chip numbers in eval.
    WIN_METRIC = NotImplementedError  # To plot graphs in relation to a fixed quantity (e.g. Poker.MeasureAnte)

    N_HOLE_CARDS = NotImplementedError  # number of private cards
    N_RANKS = NotImplementedError  # number of card-ranks
    N_SUITS = NotImplementedError  # number of card-suits
    N_CARDS_IN_DECK = NotImplementedError  # N_RANKS * N_SUITS
    RANGE_SIZE = NotImplementedError  # number of possible unique combinations of holecards

    BTN_IS_FIRST_POSTFLOP = NotImplementedError
    FIRST_ACTION_NO_CALL = False

    IS_FIXED_LIMIT_GAME = NotImplementedError
    IS_POT_LIMIT_GAME = NotImplementedError

    # Only relevant if Limit game!
    MAX_N_RAISES_PER_ROUND = NotImplementedError
    ROUND_WHERE_BIG_BET_STARTS = NotImplementedError

    # obs modes
    SUITS_MATTER = NotImplementedError  # Whether suits matter in the ruleset played. (i.e. do flushes exist?)

    N_FLOP_CARDS = NotImplementedError
    N_TURN_CARDS = NotImplementedError
    N_RIVER_CARDS = NotImplementedError
    N_TOTAL_BOARD_CARDS = NotImplementedError

    # This MUST NOT skip rounds. Setting it to [PREFLOP, FLOP] is ok, but [PREFLOP, TURN] will not work.
    ALL_ROUNDS_LIST = NotImplementedError

    # Dicts mapping rounds to other rounds.
    ROUND_BEFORE = NotImplementedError
    ROUND_AFTER = NotImplementedError

    # Map ints to strings for printing cards
    RANK_DICT = NotImplementedError
    SUIT_DICT = NotImplementedError

    # Class that contains the base-rulset the game follows
    RULES = NotImplementedError

    # ____________________________________________________ CONSTRUCT ___________________________________________________
    def __init__(self,
                 env_args,
                 lut_holder,
                 is_evaluating,
                 ):
        """
        Args:
            env_args:               Depending on game type an instance of PokerEnvArgs or DiscretePokerEnvArgs

            lut_holder:             Depending on game type. An instance of a subclass of LutHolder. It is not checked
                                    whether the correct one is passed, so be sure to pass the right one!
                                    lut_holder could theoretically be created encapsulated by an instance of this class,
                                    but for optimization (i.e. only one per machine, not per env), we pass it.

            is_evaluating (bool):   Whether the environment shall be spawned in evaluation mode (i.e. no randomization)
                                    or not.
        """
        assert env_args.n_seats >= 2

        self._args = copy.deepcopy(env_args)
        self.lut_holder = lut_holder
        self.IS_EVALUATING = is_evaluating

        # deck of cards
        self.deck = DeckOfCards(num_suits=self.N_SUITS, num_ranks=self.N_RANKS)

        # Initialize args
        self.BTN_POS = NotImplementedError
        self.SB_POS = NotImplementedError
        self.BB_POS = NotImplementedError
        self._USE_SIMPLE_HU_OBS = NotImplementedError
        self.RETURN_PRE_TRANSITION_STATE_IN_INFO = NotImplementedError
        self.N_SEATS = NotImplementedError
        self.MAX_CHIPS = NotImplementedError
        self.STACK_RANDOMIZATION_RANGE = NotImplementedError
        self.REWARD_SCALAR = NotImplementedError
        self.seats = NotImplementedError
        self._init_from_args(env_args=env_args, is_evaluating=is_evaluating)

        # ______________________________  Observation- & Action-space ______________________________
        self.observation_space, self.obs_idx_dict, self.obs_parts_idxs_dict = self._construct_obs_space()
        self.hole_card_space_shape = [self.N_HOLE_CARDS, 2]

        # __________________________________  episode state vars  __________________________________
        # all these vars should be initialized by calling env.reset() in the training script.
        self.current_round = None
        self.side_pots = None  # chip count in side pot.  list with len=n_seats
        self.main_pot = None  # chip count in main pot
        self.board = None  # np.ndarray(shape=(n_cards, 2))
        self.last_action = None  # list of 3 ints: [action_idx, _raise_amount, player.seat_id]
        self.capped_raise = CappedRaise()  # [happend_this_round, player_that_raised, player_that_cant_reopen]
        self.current_player = None  # PokerPlayer instance of player currently having to get_action
        self.last_raiser = None  # PokerPlayer instance
        self.n_actions_this_episode = None  # Number of actions performed this episode

        # only relevant in Limit games
        self.n_raises_this_round = NotImplementedError

    def _construct_obs_space(self):
        """
        The maximum all chip-values can reach is n_seats, because we normalize by dividing by the average starting stack
        """
        obs_idx_dict = {}
        obs_parts_idxs_dict = {
            "board": [],
            "players": [[] for _ in range(self.N_SEATS)],
            "table_state": [],
        }
        next_idx = [0]  # list is a mutatable object. int not.

        def get_discrete(size, name, _curr_idx):
            obs_idx_dict[name] = _curr_idx[0]
            _curr_idx[0] += 1
            return spaces.Discrete(size)

        def get_new_box(name, _curr_idx, high, low=0):
            obs_idx_dict[name] = _curr_idx[0]
            _curr_idx[0] += 1
            return spaces.Box(low=low, high=high, shape=(1,), dtype=np.float32)

        if (self.N_SEATS == 2) and self._USE_SIMPLE_HU_OBS:
            # __________________________  Public Information About Game State  _________________________
            _k = next_idx[0]
            _table_space = [  # (blinds are in obs to give the agent a perspective on starting stack after normalization
                get_new_box("ante", next_idx, self.N_SEATS),  # .................................... self.ANTE
                get_new_box("small_blind", next_idx, self.N_SEATS),  # ............................. self.SMALL_BLIND
                get_new_box("big_blind", next_idx, self.N_SEATS),  # ............................... self.BIG_BLIND
                get_new_box("min_raise", next_idx, self.N_SEATS),  # ............................... min total raise
                get_new_box("pot_amt", next_idx, self.N_SEATS),  # ................................. main_pot amount
                get_new_box("total_to_call", next_idx, self.N_SEATS),  # ........................... total_to_call
                get_new_box("last_action_how_much", next_idx, self.N_SEATS),  # .................... self.last_action[1]
            ]
            for i in range(3):  # .................................................................. self.last_action[0]
                _table_space.append(get_discrete(1, "last_action_what_" + str(i), next_idx))

            for i in range(
                self.N_SEATS):  # ....................................................... self.last_action[2]
                _table_space.append(get_discrete(1, "last_action_who_" + str(i), next_idx))

            for i in range(
                self.N_SEATS):  # ....................................................... curr_player.seat_id
                _table_space.append(get_discrete(1, "p" + str(i) + "_acts_next", next_idx))

            for i in range(max(self.ALL_ROUNDS_LIST) + 1):  # ...................................... round onehot
                _table_space.append(get_discrete(1, "round_" + Poker.INT2STRING_ROUND[i], next_idx)),

            # add to parts_dict for possible slicing for agents.
            obs_parts_idxs_dict["table_state"] += list(range(_k, next_idx[0]))

            # __________________________  Public Information About Each Player  ________________________
            _player_space = []
            for i in range(self.N_SEATS):
                _k = next_idx[0]
                _player_space += [
                    get_new_box("stack_p" + str(i), next_idx, self.N_SEATS),  # ..................... stack
                    get_new_box("curr_bet_p" + str(i), next_idx, self.N_SEATS),  # .................. current_bet
                    get_discrete(1, "is_allin_p" + str(i), next_idx),  # ............................ is_allin
                ]

                # add to parts_dict for possible slicing for agents
                obs_parts_idxs_dict["players"][i] += list(range(_k, next_idx[0]))

            # _______________________________  Public cards (i.e. board)  ______________________________
            _board_space = []
            _k = next_idx[0]
            for i in range(self.N_TOTAL_BOARD_CARDS):

                x = []
                for j in range(self.N_RANKS):  # .................................................... rank
                    x.append(get_discrete(1, str(i) + "th_board_card_rank_" + str(j), next_idx))

                for j in range(self.N_SUITS):  # .................................................... suit
                    x.append(get_discrete(1, str(i) + "th_board_card_suit_" + str(j), next_idx))

                _board_space += x

            # add to parts_dict for possible slicing for agents
            obs_parts_idxs_dict["board"] += list(range(_k, next_idx[0]))

            # __________________________  Return Complete _Observation Space  __________________________
            # Tuple (lots of spaces.Discrete and spaces.Box)
            _observation_space = spaces.Tuple(_table_space + _player_space + _board_space)
            _observation_space.shape = [len(_observation_space.spaces)]

        else:
            # __________________________  Public Information About Game State  _________________________
            _k = next_idx[0]
            _table_space = [  # (blinds are in obs to give the agent a perspective on starting stack after normalization
                get_new_box("ante", next_idx, self.N_SEATS),  # .................................... self.ANTE
                get_new_box("small_blind", next_idx, self.N_SEATS),  # ............................. self.SMALL_BLIND
                get_new_box("big_blind", next_idx, self.N_SEATS),  # ............................... self.BIG_BLIND
                get_new_box("min_raise", next_idx, self.N_SEATS),  # ............................... min total raise
                get_new_box("pot_amt", next_idx, self.N_SEATS),  # ................................. main_pot amount
                get_new_box("total_to_call", next_idx, self.N_SEATS),  # ........................... total_to_call
                get_new_box("last_action_how_much", next_idx, self.N_SEATS),  # .................... self.last_action[1]
            ]
            for i in range(3):  # .................................................................. self.last_action[0]
                _table_space.append(get_discrete(1, "last_action_what_" + str(i), next_idx))

            for i in range(self.N_SEATS):  # ....................................................... self.last_action[2]
                _table_space.append(get_discrete(1, "last_action_who_" + str(i), next_idx))

            for i in range(self.N_SEATS):  # ....................................................... curr_player.seat_id
                _table_space.append(get_discrete(1, "p" + str(i) + "_acts_next", next_idx))

            for i in range(max(self.ALL_ROUNDS_LIST) + 1):  # ...................................... round onehot
                _table_space.append(get_discrete(1, "round_" + Poker.INT2STRING_ROUND[i], next_idx)),

            for i in range(self.N_SEATS):  # ....................................................... side pots
                _table_space.append(get_new_box("side_pot_" + str(i), next_idx, 1))

            # add to parts_dict for possible slicing for agents.
            obs_parts_idxs_dict["table_state"] += list(range(_k, next_idx[0]))

            # __________________________  Public Information About Each Player  ________________________
            _player_space = []
            for i in range(self.N_SEATS):
                _k = next_idx[0]
                _player_space += [
                    get_new_box("stack_p" + str(i), next_idx, self.N_SEATS),  # ..................... stack
                    get_new_box("curr_bet_p" + str(i), next_idx, self.N_SEATS),  # .................. current_bet
                    get_discrete(1, "has_folded_this_episode_p" + str(i), next_idx),  # ............. folded_this_epis
                    get_discrete(1, "is_allin_p" + str(i), next_idx),  # ............................ is_allin
                ]
                for j in range(self.N_SEATS):
                    _player_space.append(
                        get_discrete(1, "side_pot_rank_p" + str(i) + "_is_" + str(j), next_idx))  # . side_pot_rank

                # add to parts_dict for possible slicing for agents
                obs_parts_idxs_dict["players"][i] += list(range(_k, next_idx[0]))

            # _______________________________  Public cards (i.e. board)  ______________________________
            _board_space = []
            _k = next_idx[0]
            for i in range(self.N_TOTAL_BOARD_CARDS):

                x = []
                for j in range(self.N_RANKS):  # .................................................... rank
                    x.append(get_discrete(1, str(i) + "th_board_card_rank_" + str(j), next_idx))

                for j in range(self.N_SUITS):  # .................................................... suit
                    x.append(get_discrete(1, str(i) + "th_board_card_suit_" + str(j), next_idx))

                _board_space += x

            # add to parts_dict for possible slicing for agents
            obs_parts_idxs_dict["board"] += list(range(_k, next_idx[0]))

            # __________________________  Return Complete _Observation Space  __________________________
            # Tuple (lots of spaces.Discrete and spaces.Box)
            _observation_space = spaces.Tuple(_table_space + _player_space + _board_space)
            _observation_space.shape = [len(_observation_space.spaces)]
        return _observation_space, obs_idx_dict, obs_parts_idxs_dict

    def _init_from_args(self, env_args, is_evaluating):
        a = copy.deepcopy(env_args)

        # Heads-Up rules
        if a.n_seats == 2:
            self.BTN_POS = 0
            self.SB_POS = 0
            self.BB_POS = 1

        # >2 player poker rules
        else:
            self.BTN_POS = 0
            self.SB_POS = 1
            self.BB_POS = 2

        self._USE_SIMPLE_HU_OBS = a.use_simplified_headsup_obs
        self.RETURN_PRE_TRANSITION_STATE_IN_INFO = a.RETURN_PRE_TRANSITION_STATE_IN_INFO
        self.N_SEATS = int(a.n_seats)

        try:
            self.MAX_CHIPS = sum(a.starting_stack_sizes_list) \
                             + a.stack_randomization_range[1] * a.n_seats \
                             + 1
        except TypeError:  # stack size set to None -> Default
            self.MAX_CHIPS = a.n_seats * (self.DEFAULT_STACK_SIZE + a.stack_randomization_range[1]) + 1

        self.STACK_RANDOMIZATION_RANGE = a.stack_randomization_range

        if a.scale_rewards:
            try:
                self.REWARD_SCALAR = float(sum(a.starting_stack_sizes_list)) / float(a.n_seats) / 5
            except TypeError:  # stack size set to None -> Default
                self.REWARD_SCALAR = self.DEFAULT_STACK_SIZE / 5.0

        else:
            self.REWARD_SCALAR = 1.0

        # fill seats with players
        self.seats = [
            PokerPlayer(seat_id=i,
                        poker_env=self,
                        is_evaluating=is_evaluating,
                        starting_stack=
                        (a.starting_stack_sizes_list[i]
                         if a.starting_stack_sizes_list[i] is not None
                         else self.DEFAULT_STACK_SIZE),
                        stack_randomization_range=a.stack_randomization_range)
            for i in range(a.n_seats)]

    # __________________________________________________ TO OVERRIDE ___________________________________________________
    def get_hand_rank(self, hand_2d, board_2d):
        """

        Args:
            hand_2d (np.ndarray):       the hand to evaluate
            board_2d (np.ndarray):      the board to evaluate

        Returns:
            int: handrank, higher is better.

        """
        raise NotImplementedError

    def get_hand_rank_all_hands_on_given_boards(self, boards_1d, lut_holder):
        """
        This function allows computing the hand rank for all possible hands on multiple boards at once.

        Args:
            boards_1d (np.ndarray):     an array of 1D board representations
            lut_holder:                 a LUT associated with this type of environment.
        """
        raise NotImplementedError

    def _get_env_adjusted_action_formulation(self, action):
        """
        A Discretized PokerEnv subclass has a different action representation than a Limit or a No-Limit env.
        This function should be used to convert these different action spaces to the standard one of PokerEnv:
        Tuple(Discrete(3), Discrete(n_chips))
        Or in words: (action_to_take, n_chips_to_bet_if_action_is_bet).
        n_chips_to_bet_if_action_is_bet always needs to be set but is ignored if the action is not Poker.BET_RAISE

        Args:
            action: subclass specific action representation

        Returns:
            Tuple:  (action_to_take, n_chips_to_bet_if_action_is_bet)
        """
        return action

    def _adjust_raise(self, raise_total_amount_in_chips):
        """
        Different per game type
        Args:
            raise_total_amount_in_chips: chips intended to raise

        Returns:
            Chips to raise to according to gametype and ""raise_total_amount_in_chips"".
        """
        raise NotImplementedError

    @staticmethod
    def get_lut_holder():
        """ return an instance of a lutholder specific to the game's rules"""
        raise NotImplementedError

    # _____________________________________________________ POKER ______________________________________________________
    def _deal_hole_cards(self):
        for player in self.seats:
            player.hand = self.deck.draw(self.N_HOLE_CARDS)

    def _deal_flop(self):
        self.board[:self.N_FLOP_CARDS] = self.deck.draw(self.N_FLOP_CARDS)

    def _deal_turn(self):
        self.board[self.N_FLOP_CARDS:self.N_FLOP_CARDS + self.N_TURN_CARDS] = self.deck.draw(self.N_TURN_CARDS)

    def _deal_river(self):
        d = self.N_FLOP_CARDS + self.N_TURN_CARDS
        self.board[d:d + self.N_RIVER_CARDS] = self.deck.draw(self.N_RIVER_CARDS)

    def _post_antes(self):
        for s in self.seats:
            s.bet_raise(self.ANTE)
            s.has_acted_this_round = False

    def _post_small_blind(self):
        player = self.seats[self.SB_POS]
        player.bet_raise(self.SMALL_BLIND)
        player.has_acted_this_round = False

    def _post_big_blind(self):
        player = self.seats[self.BB_POS]
        player.bet_raise(self.BIG_BLIND)
        player.has_acted_this_round = False

    def _payout_pots(self):
        self._assign_hand_ranks_to_all_players()

        if self.N_SEATS == 2:
            if self.seats[0].hand_rank > self.seats[1].hand_rank:
                self.seats[0].award(self.main_pot)
            elif self.seats[0].hand_rank < self.seats[1].hand_rank:
                self.seats[1].award(self.main_pot)
            else:
                # in HU the number of chips is always even because both players had to put the same amount in.
                self.seats[0].award(self.main_pot / 2)
                self.seats[1].award(self.main_pot / 2)

            self.main_pot = 0

        else:
            pots = np.array([self.main_pot] + self.side_pots)  # appends mainpot as idx 0 to sidepots
            pot_ranks = np.arange(start=-1, stop=len(self.side_pots))
            pot_and_pot_ranks = np.array((pots, pot_ranks)).T

            for e in pot_and_pot_ranks:
                pot = e[0]
                rank = e[1]
                eligible_players = [p for p in self.seats if p.side_pot_rank >= rank and not p.folded_this_episode]

                num_eligible = len(eligible_players)
                if num_eligible > 0:

                    # side_pot / num_winners could be non-int! if so, we decide randomly who to give the chip to.
                    winner_list = self._get_winner_list(players_to_consider=eligible_players)  # PokerPlayer objects
                    num_winners = int(len(winner_list))

                    chips_per_winner = int(pot / num_winners)  # rounds down
                    num_non_div_chips = int(pot) % num_winners  # distribute afterwards

                    for p in winner_list:
                        p.award(chips_per_winner)

                    # distribute the rest randomly.
                    shuffled_winner_idxs = np.arange(num_winners)
                    np.random.shuffle(shuffled_winner_idxs)
                    for p_idx in shuffled_winner_idxs[:num_non_div_chips]:
                        self.seats[p_idx].award(1)

            # set all 0
            self.side_pots = [0] * self.N_SEATS
            self.main_pot = 0

    def _pay_all_to_one_player(self, player_to_pay_to):
        """
        IGNORES SIDEPOT RANKS

        Args:
            player_to_pay_to (PokerPlayer)
        """
        for seat in self.seats:
            player_to_pay_to.award(seat.current_bet)
            seat.current_bet = 0

        player_to_pay_to.award(sum(self.side_pots))
        self.side_pots = [0] * self.N_SEATS

        player_to_pay_to.award(self.main_pot)
        self.main_pot = 0

    def _assign_hand_ranks_to_all_players(self):
        for player in self.seats:
            player.hand_rank = self.get_hand_rank(hand_2d=player.hand, board_2d=self.board)

    def _put_current_bets_into_main_pot_and_side_pots(self):

        if self.N_SEATS == 2:
            # If there is a difference between biggest and 2nd biggest current_bet, award that to the player.
            dif_p0_to_p1 = self.seats[0].current_bet - self.seats[1].current_bet

            if dif_p0_to_p1 > 0:
                self.seats[0].refund_from_bet(dif_p0_to_p1)
            elif dif_p0_to_p1 < 0:
                self.seats[1].refund_from_bet(-dif_p0_to_p1)

            self.main_pot += self.seats[0].current_bet
            self.main_pot += self.seats[1].current_bet
            self.seats[0].current_bet = 0
            self.seats[1].current_bet = 0

        else:
            # If there is a difference between biggest and 2nd biggest current_bet, award that to the player.
            _players_sorted_by_bet_in_front = sorted(self.seats, key=lambda x: x.current_bet, reverse=True)

            dif = _players_sorted_by_bet_in_front[0].current_bet - _players_sorted_by_bet_in_front[1].current_bet
            _players_sorted_by_bet_in_front[0].refund_from_bet(dif)

            # ________________________________________ fill main_pot ________________________________________
            players_not_folded = [p for p in self.seats if not p.folded_this_episode]

            # The smallest bet of someone who did not fold.
            # If no one is allin, all nonfold bets are equal and put into main pot
            main_pot_max_amount = min([p.current_bet for p in players_not_folded])

            for p in self.seats:
                # if bigger needs sidepot. if smaller means he folded at some point and just did not contribute as much
                amount_contributed = min(p.current_bet, main_pot_max_amount)

                self.main_pot += amount_contributed
                p.current_bet -= amount_contributed

            # _____________________________________ maybe fill side pots ______________________________________
            def _find_next_smallest_bet():
                """
                Returns:
                    if all current_bets are 0: None
                    otherwise: idx of smallest nonzero bet

                """
                current_bets = [p.current_bet for p in self.seats]
                next_bet_idx = None

                for b_idx in range(self.N_SEATS):
                    if current_bets[b_idx] > 0:

                        # and not self.seats[b_idx].folded_this_episode   is safe because a folded player can't
                        if ((next_bet_idx is None or current_bets[b_idx] < current_bets[next_bet_idx])
                            and not self.seats[b_idx].folded_this_episode):
                            next_bet_idx = b_idx

                return next_bet_idx

            idx_smallest_bet = _find_next_smallest_bet()  # get status after main pot calc

            while idx_smallest_bet is not None:
                current_max_side_pot_rank = max([p.side_pot_rank for p in self.seats])  # -1 on default.
                side_pot_idx = current_max_side_pot_rank + 1

                side_pot_amount_per_player_in_it = self.seats[idx_smallest_bet].current_bet

                players_not_all_in_after_this_cleanup = [p for p in self.seats if not (
                    p.current_bet < side_pot_amount_per_player_in_it and p.is_allin)]

                for p in players_not_all_in_after_this_cleanup:
                    p.side_pot_rank = side_pot_idx

                # fill this sidepot
                for p in self.seats:
                    # if bigger: needs another sidepot
                    # if smaller: means he folded at some point and just did not contribute as much
                    amount_contributed = min(p.current_bet, side_pot_amount_per_player_in_it)
                    self.side_pots[side_pot_idx] += amount_contributed
                    p.current_bet -= amount_contributed

                # for next iteration
                idx_smallest_bet = _find_next_smallest_bet()  # get status after main pot calc

    def _rundown(self):
        while True:
            self.current_round += 1

            if self.current_round == self.ALL_ROUNDS_LIST[-1] + 1:
                self._put_current_bets_into_main_pot_and_side_pots()
                self.current_round -= 1  # we still want to be on the River.

                if self.RETURN_PRE_TRANSITION_STATE_IN_INFO:
                    state_before_transition = self.state_dict()

                self._payout_pots()

                if self.RETURN_PRE_TRANSITION_STATE_IN_INFO:
                    return state_before_transition
                return

            elif self.current_round == Poker.FLOP:
                self._deal_flop()
            elif self.current_round == Poker.TURN:
                self._deal_turn()
            elif self.current_round == Poker.RIVER:
                self._deal_river()
            else:
                raise ValueError(self.current_round)

    def _deal_next_round(self):
        """
        Call this AFTER round+=1
        """
        if self.current_round == Poker.PREFLOP:
            self._deal_hole_cards()
        elif self.current_round == Poker.FLOP:
            self._deal_flop()
        elif self.current_round == Poker.TURN:
            self._deal_turn()
        elif self.current_round == Poker.RIVER:
            self._deal_river()
        else:
            raise ValueError(self.current_round)

    def _next_round(self):
        if self.IS_FIXED_LIMIT_GAME:
            self.n_raises_this_round = 0

        # refer to #ID_2 in docstring of this class for this.
        self.capped_raise.reset()

        # sort out mainpot, sidepots and p.currentbets
        self._put_current_bets_into_main_pot_and_side_pots()

        # This must be called BEFORE round += 1
        self.current_player = self._get_first_to_act_post_flop()

        # set has_acted_this_round = False and maybe others
        for p in self.seats:
            p.has_acted_this_round = False

        self.current_round += 1  # highly dependant on PokerEnv.""ROUND_NAME"" being sequential ints!
        self._deal_next_round()

    def _step(self, processed_action):
        """
        the action passed is considered to be for self.current_player and come from its respective agent (if applicable
        to your algorithm).

        actions are always of the form [action_idx, _raise_size]
        However _raise_size is only considered when the action is Poker.BET_RAISE
        raise_size is measured in total chips as current_bet. Not as an addition to current_bet

        Args:
            processed_action (tuple or list): (action_idx, raise_size)

        Returns:
            obs, rew_for_all_players, done?, info
        """

        # After this call, this fn assumes that executing the action is legal.
        processed_action = self._get_fixed_action(action=processed_action)

        if processed_action[0] == Poker.CHECK_CALL:
            self.current_player.check_call(total_to_call=processed_action[1])

        elif processed_action[0] == Poker.FOLD:
            self.current_player.fold()

        elif processed_action[0] == Poker.BET_RAISE:

            # This happens when someone has fewer chips than minraise and goes all in.
            # The last raiser, if there is one, can't reopen in this case until someone else reraises!
            if processed_action[1] < self._get_current_total_min_raise():
                self.capped_raise.happened_this_round = True
                self.capped_raise.player_that_raised = self.current_player
                self.capped_raise.player_that_cant_reopen = self.last_raiser  # might be None. Then everyone can raise.

            elif self.capped_raise.happened_this_round is True:
                # if someone shoved under minraise over someone else's raise the original raiser can't reraise again!
                # But if a 3rd plyr raises over that under-min shove, everyone can raise again. this is handled here.
                if self.capped_raise.player_that_cant_reopen is not self.current_player:
                    self.capped_raise.reset()

            self.last_raiser = self.current_player  # leave this line at the end of this function!!
            self.current_player.bet_raise(total_bet_amount=processed_action[1])

            self.n_actions_this_episode += 1

            # If this is a limit poker game, increment the raise counter
            if self.IS_FIXED_LIMIT_GAME:
                self.n_raises_this_round += 1
        else:
            raise RuntimeError(processed_action[0], " is not legal")

        self.last_action = [processed_action[0], processed_action[1], self.current_player.seat_id]

        # ______________________________________________________________________________________________________________
        # check if should deal next round, rundown or continue to next step in the episode of the poker game

        all_non_all_in_and_non_fold_p = [p for p in self.seats if not p.folded_this_episode and not p.is_allin]
        all_nonfold_p = [p for p in self.seats if not p.folded_this_episode]

        # just let next player run in this round
        info = None
        if self._should_continue_in_this_round(all_non_all_in_and_non_fold_p=all_non_all_in_and_non_fold_p,
                                               all_nonfold_p=all_nonfold_p):
            self.current_player = self._get_player_that_has_to_act_next()
            is_terminal = False

            if self.RETURN_PRE_TRANSITION_STATE_IN_INFO:
                info = {"chance_acts": False, "state_dict_before_money_move": None}

        # next round
        elif len(all_non_all_in_and_non_fold_p) > 1:

            # payout if final round
            if self.current_round == len(self.ALL_ROUNDS_LIST) - 1:
                is_terminal = True
                self._put_current_bets_into_main_pot_and_side_pots()
                if self.RETURN_PRE_TRANSITION_STATE_IN_INFO:
                    info = {"chance_acts": False, "state_dict_before_money_move": self.state_dict()}
                self._payout_pots()

            # deal next round
            else:
                is_terminal = False
                if self.RETURN_PRE_TRANSITION_STATE_IN_INFO:
                    info = {"chance_acts": True, "state_dict_before_money_move": self.state_dict()}
                self._next_round()

        # rundown
        elif len(all_nonfold_p) > 1:  # rundown only makes sense if >0 are allin and 1 is not or >2 are allin.
            is_terminal = True
            state_before_payouts = self._rundown()

            if self.RETURN_PRE_TRANSITION_STATE_IN_INFO:
                info = {"chance_acts": False, "state_dict_before_money_move": state_before_payouts}

        # only one not folded, so pay all pots to him.
        elif len(all_nonfold_p) == 1:
            is_terminal = True
            if self.RETURN_PRE_TRANSITION_STATE_IN_INFO:
                self._put_current_bets_into_main_pot_and_side_pots()
                info = {"chance_acts": False, "state_dict_before_money_move": self.state_dict()}
                self._payout_pots()
            else:  # more efficient, but doesnt give info needed.
                self._pay_all_to_one_player(all_nonfold_p[0])

        else:
            raise RuntimeError("There seems to be an edge-case not built into this")

        return self._get_current_step_returns(is_terminal=is_terminal, info=info)

    # _____________________________________________________ UTIL  ______________________________________________________
    def _get_winner_list(self, players_to_consider):
        """
        Returns:
            list: list of PokerPlayer instances that are winners

        """
        best_rank = max([p.hand_rank for p in players_to_consider])
        winners = [p for p in players_to_consider if p.hand_rank == best_rank]

        return winners

    def _get_current_total_min_raise(self):
        """
        Taking the highest and 2nd highest and subtracting them gives us the minraise amount. If all bets are equal,
        we return with a delta of 1 big blind.
        """

        if self.N_SEATS == 2:
            _sorted_ascending = sorted([p.current_bet for p in self.seats])  # 0 is small, 1 is big
            delta = max(_sorted_ascending[1] - _sorted_ascending[0], self.BIG_BLIND)
            return _sorted_ascending[1] + delta

        else:
            current_bets_sorted_descending = sorted([p.current_bet for p in self.seats], reverse=True)
            current_to_call_total = max(current_bets_sorted_descending)
            _largest_bet = current_bets_sorted_descending[0]

            for i in range(1, self.N_SEATS):
                if current_bets_sorted_descending[i] == _largest_bet:
                    continue

                delta_between_last_and_before_last = _largest_bet - current_bets_sorted_descending[i]

                delta = max(delta_between_last_and_before_last, self.BIG_BLIND)
                return current_to_call_total + delta

            # in cases where all bets are equal, the minraise delta is 1 big blind
            return current_to_call_total + self.BIG_BLIND

    def _get_new_board(self):
        return np.full((self.N_TOTAL_BOARD_CARDS, 2), Poker.CARD_NOT_DEALT_TOKEN_1D, dtype=np.int8)

    def _get_first_to_act_pre_flop(self):
        if self.N_SEATS >= 4:
            # left of BB
            return self.seats[3]
        else:
            # for n_players==3 and 2 btn starts
            return self.seats[0]

    def _get_first_to_act_post_flop(self):
        """
        Btn has index 0. He is always the last with the exception of some HU rules where ""BTN_IS_FIRST_POSTFLOP"" can
        be set to True. In multi-agent games, we search for the smalles seat id in the list, while 0 (i.e. btn) is
        treated as inf.
        """

        if self.N_SEATS == 2:
            if self.BTN_IS_FIRST_POSTFLOP:
                return self.seats[0]
            else:
                return self.seats[1]
        else:
            players_to_consider = [p for p in self.seats if not p.folded_this_episode and not p.is_allin]

            # since there will always be at least 2 ppl in the pot, the btn can NEVER be the first!
            first_p = players_to_consider[0]
            for p in players_to_consider:
                if p.seat_id < first_p.seat_id or first_p.seat_id == 0:
                    first_p = p

            return first_p

    def _get_biggest_bet_out_there_aka_total_to_call(self):
        """
        chip count of max([p.current_bet for p in self.seats])
        """
        return max([p.current_bet for p in self.seats])

    def _get_player_that_has_to_act_next(self):
        idx = self.seats.index(self.current_player) + 1

        for i in range(self.N_SEATS):
            mod_idx = idx % self.N_SEATS
            p = self.seats[mod_idx]

            if not p.is_allin and not p.folded_this_episode:
                return self.seats[mod_idx]

            idx += 1

        raise RuntimeError("There is no next player. Seems like some more debugging is needed...")

    def _get_fixed_action(self, action):
        """
        This method is responsible for asserting that an action is valid at the current state and returns the
        capped/changed action if not.

        Args:
            action (iterable):      iterable of 2 ints - [PokerEnv.ACTIONTYPE, _raise_amount_in_chips]

        Returns:            [Poker.FOLD, -1]
                        or  [Poker.CHECK_CALL, total_bet_to_be_placed_in_front_by_player]
                        or  [Poker.BET_RAISE, total_bet_to_be_placed_in_front_by_player]

        """
        _action_idx = action[0]

        total_to_call = self._get_biggest_bet_out_there_aka_total_to_call()

        if _action_idx == Poker.FOLD:
            if total_to_call <= self.current_player.current_bet:
                return self._process_check_call(total_to_call=total_to_call)
            else:
                return [Poker.FOLD, -1]

        elif _action_idx == Poker.CHECK_CALL:
            if (self.FIRST_ACTION_NO_CALL
                and (self.n_actions_this_episode == 0)
                and self.current_round == Poker.PREFLOP):
                return [Poker.FOLD, -1]

            return self._process_check_call(total_to_call=total_to_call)

        elif _action_idx == Poker.BET_RAISE:

            # Limit Poker specific rule
            if self.IS_FIXED_LIMIT_GAME:
                if self.n_raises_this_round >= self.MAX_N_RAISES_PER_ROUND[self.current_round]:
                    return self._process_check_call(total_to_call=total_to_call)

            if ((self.current_player.stack + self.current_player.current_bet <= total_to_call)
                or (self.capped_raise.player_that_cant_reopen is self.current_player)):
                return self._process_check_call(total_to_call=total_to_call)
            else:
                return self._process_raise(raise_total_amount_in_chips=action[1])
        else:
            raise RuntimeError('invalid action ({}) must be fold (0), call (1), or raise (2) '.format(_action_idx))

    def _process_check_call(self, total_to_call):
        delta_to_call = min(total_to_call - self.current_player.current_bet, self.current_player.stack)
        total_bet_to_be_placed = int(delta_to_call + self.current_player.current_bet)
        return [Poker.CHECK_CALL, total_bet_to_be_placed]

    def _process_raise(self, raise_total_amount_in_chips):
        raise_to = self._adjust_raise(raise_total_amount_in_chips=raise_total_amount_in_chips)
        # lastly, if that amount is too much, raise all in
        if self.current_player.current_bet + self.current_player.stack < raise_to:
            raise_to = self.current_player.stack + self.current_player.current_bet
        return [Poker.BET_RAISE, int(raise_to)]

    def _should_continue_in_this_round(self, all_non_all_in_and_non_fold_p, all_nonfold_p):
        """ util function used in ._step() """

        # only 1 player did not fold yet
        if len(all_nonfold_p) < 2:
            return False

        largest_bet = max([p.current_bet for p in self.seats])
        if len([p for p in all_nonfold_p if p.is_allin or p.current_bet == largest_bet]) == len(all_nonfold_p) \
            and len([p for p in all_non_all_in_and_non_fold_p if not p.has_acted_this_round]) == 0:
            return False
        return True

    # _____________________________________________________ OUTPUT _____________________________________________________
    def _get_current_step_returns(self, is_terminal, info):
        obs = self.get_current_obs(is_terminal)
        reward = self._get_step_reward(is_terminal)
        return obs, reward, is_terminal, info

    def _get_player_states_all_players(self, normalization_sum):
        """ Public Information About Each Player """
        if (self.N_SEATS == 2) and self._USE_SIMPLE_HU_OBS:
            player_states = []
            for player in self.seats:
                player_states += [
                    player.stack / normalization_sum,
                    player.current_bet / normalization_sum,
                    player.is_allin
                ]

        else:
            player_states = []
            for player in self.seats:
                player_states += [
                    player.stack / normalization_sum,
                    player.current_bet / normalization_sum,
                    player.folded_this_episode,
                    player.is_allin
                ]
                x = [0] * self.N_SEATS
                if player.side_pot_rank > 0:
                    x[int(player.side_pot_rank)] = 1
                player_states += x

        return player_states

    def _get_board_state(self, ):
        K = (self.N_RANKS + self.N_SUITS)
        _board_space = [0] * (self.N_TOTAL_BOARD_CARDS * K)
        for i, card in enumerate(self.board.tolist()):
            if card[0] == Poker.CARD_NOT_DEALT_TOKEN_1D:
                break
            D = K * i
            _board_space[card[0] + D] = 1

            # only include suit in obs if it is relevant to the game rules played
            if self.SUITS_MATTER:
                _board_space[card[1] + D + self.N_RANKS] = 1

        return _board_space

    def _get_table_state(self, normalization_sum):
        if (self.N_SEATS == 2) and self._USE_SIMPLE_HU_OBS:
            community_state = [
                self.ANTE / normalization_sum,
                self.SMALL_BLIND / normalization_sum,
                self.BIG_BLIND / normalization_sum,
                self._get_current_total_min_raise() / normalization_sum,
                self.main_pot / normalization_sum,
                self._get_biggest_bet_out_there_aka_total_to_call() / normalization_sum,
                self.last_action[1] / normalization_sum if self.last_action[0] is not None else 0,
            ]

            x_what = [0] * 3  # last action what
            x_who = [0] * self.N_SEATS  # last action who
            if self.last_action[0] is not None:
                x_who[self.last_action[2]] = 1
                x_what[self.last_action[0]] = 1
            community_state += x_what + x_who

            # who acts next?
            x = [0] * self.N_SEATS
            x[self.current_player.seat_id] = 1
            community_state += x

            # current round
            x = [0] * (self.ALL_ROUNDS_LIST[-1] + 1)
            x[self.current_round] = 1
            community_state += x

        else:
            community_state = [
                self.ANTE / normalization_sum,
                self.SMALL_BLIND / normalization_sum,
                self.BIG_BLIND / normalization_sum,
                self._get_current_total_min_raise() / normalization_sum,
                self.main_pot / normalization_sum,
                self._get_biggest_bet_out_there_aka_total_to_call() / normalization_sum,
                self.last_action[1] / normalization_sum if self.last_action[0] is not None else 0,
            ]

            x_what = [0] * 3  # last action what
            x_who = [0] * self.N_SEATS  # last action who
            if self.last_action[0] is not None:
                x_who[self.last_action[2]] = 1
                x_what[self.last_action[0]] = 1
            community_state += x_what + x_who

            # who acts next?
            x = [0] * self.N_SEATS
            x[self.current_player.seat_id] = 1
            community_state += x

            # current round
            x = [0] * (self.ALL_ROUNDS_LIST[-1] + 1)
            x[self.current_round] = 1
            community_state += x

            # side_pots
            if self.N_SEATS > 2:
                community_state += [sp / normalization_sum for sp in self.side_pots]
            else:
                community_state += [0] * self.N_SEATS

        return community_state

    def _get_step_reward(self, is_terminal):
        if not is_terminal:
            return np.zeros(shape=self.N_SEATS, dtype=np.float32)
        return [(p.stack - p.starting_stack_this_episode) / self.REWARD_SCALAR for p in self.seats]

    # ______________________________________________________ API _______________________________________________________
    def reset(self, deck_state_dict=None):
        """
        Resets the state of the game to the standard beginning of the episode. If specified in the args passed,
        stack size randomization is applied in the new episode. If deck_state_dict is not None, the cards
        and associated random variables are synchronized FROM the given environment, so that when .step() is called on
        each of them, they produce the same result.

        Args:
            deck_state_dict:       Optional.
                                            If an instance of a PokerEnv subclass is passed, the deck, holecards, and
                                            board in this instance will be synchronized from the handed env cls.
        """
        if self.IS_FIXED_LIMIT_GAME:
            if self.BIG_BLIND > 0:
                self.n_raises_this_round = 1  # big blind counts, but in ante-only games like LEDUC it doesn't count
            else:
                self.n_raises_this_round = 0

        # reset table
        self.side_pots = [0] * self.N_SEATS  # chip count in side pots
        self.main_pot = 0  # chip count in main pot
        self.board = self._get_new_board()
        self.last_action = [None, None, None]  # list of 3 ints: [action_idx, _raise_amount, player.seat_id]
        self.current_round = self.ALL_ROUNDS_LIST[0]
        self.capped_raise.reset()
        self.last_raiser = None
        self.n_actions_this_episode = 0

        # players
        for p in self.seats:
            p.reset()

        # reset deck
        self.deck.reset()

        # start new game
        self._post_antes()
        self._put_current_bets_into_main_pot_and_side_pots()  # antes don't count to current bet
        self._post_small_blind()
        self._post_big_blind()
        self.current_player = self._get_first_to_act_pre_flop()
        self._deal_next_round()

        # optionally synchronize random variables from another env
        if deck_state_dict is not None:
            self.load_cards_state_dict(cards_state_dict=deck_state_dict)

        return self._get_current_step_returns(is_terminal=False, info=[False, None])

    def step_raise_pot_frac(self, pot_frac):
        """
        This fn is only useful to call if current_player wants to raise. Therefore it assumes that's the case.

        Args:
            pot_frac: fraction of current pot to bet/raise

        Returns:
            obs, rew_for_all_players, done?, info:

        """
        processed_action = (2, self.get_fraction_of_pot_raise(fraction=pot_frac, player_that_bets=self.current_player))
        return self._step(processed_action=processed_action)

    def step_from_processed_tuple(self, action):
        """
        Args:
            action:  processed_action (tuple or list): (action_idx, raise_size)

        Returns:
            obs, rew_for_all_players, done?, info
        """
        return self._step(action)

    def step(self, action):
        """
        Args:
            action: env specific action representation as documented in PokerEnv.

        Returns:
            obs, rew_for_all_players, done?, info
        """

        # cap min/max raises and format action. after that the action is legal 100% and can be executed blindly
        processed_action = self._get_env_adjusted_action_formulation(action)
        return self._step(processed_action=processed_action)

    def state_dict(self):
        env_state_dict = {
            EnvDictIdxs.is_evaluating: self.IS_EVALUATING,
            EnvDictIdxs.current_round: self.current_round,  # int by value
            EnvDictIdxs.side_pots: copy.deepcopy(self.side_pots),  # np array
            EnvDictIdxs.main_pot: self.main_pot,  # int by value
            EnvDictIdxs.board_2d: np.copy(self.board),  # np array
            EnvDictIdxs.last_action: copy.deepcopy(self.last_action),
            EnvDictIdxs.capped_raise: [self.capped_raise.player_that_raised.seat_id,
                                       None
                                       if self.capped_raise.player_that_cant_reopen is None
                                       else
                                       self.capped_raise.player_that_cant_reopen.seat_id]
            if self.capped_raise.happened_this_round else None,
            EnvDictIdxs.current_player: self.current_player.seat_id,  # idx in _env.seats
            EnvDictIdxs.last_raiser: None if self.last_raiser is None else self.last_raiser.seat_id,
            # idx in _env.seats
            EnvDictIdxs.deck: self.deck.state_dict(),  # np array
            EnvDictIdxs.n_actions_this_episode: self.n_actions_this_episode,  # int
            EnvDictIdxs.seats:
                [
                    {
                        PlayerDictIdxs.seat_id: p.seat_id,  # int
                        PlayerDictIdxs.hand: np.copy(p.hand),  # np array
                        PlayerDictIdxs.hand_rank: p.hand_rank,  # int by value
                        PlayerDictIdxs.stack: p.stack,  # int by value
                        PlayerDictIdxs.current_bet: p.current_bet,  # int by value
                        PlayerDictIdxs.is_allin: p.is_allin,  # bool by value
                        PlayerDictIdxs.folded_this_episode: p.folded_this_episode,  # bool by value
                        PlayerDictIdxs.has_acted_this_round: p.has_acted_this_round,  # bool by value
                        PlayerDictIdxs.side_pot_rank: p.side_pot_rank  # int by value
                    }
                    for p in self.seats]
        }
        if self.IS_FIXED_LIMIT_GAME:
            env_state_dict[EnvDictIdxs.n_raises_this_round] = self.n_raises_this_round
        return env_state_dict

    def load_state_dict(self, env_state_dict, blank_private_info=False):
        """

        Args:
            env_state_dict:
            blank_private_info (bool): If true, hole cards are going to be set to None. This is useful when loading a
                                        public state

        Returns:

        """
        self.IS_EVALUATING = env_state_dict[EnvDictIdxs.is_evaluating]
        self.current_round = env_state_dict[EnvDictIdxs.current_round]
        self.side_pots = copy.deepcopy(env_state_dict[EnvDictIdxs.side_pots])
        self.main_pot = env_state_dict[EnvDictIdxs.main_pot]
        self.board = np.copy(env_state_dict[EnvDictIdxs.board_2d])
        self.last_action = copy.deepcopy(env_state_dict[EnvDictIdxs.last_action])

        x = env_state_dict[EnvDictIdxs.capped_raise]
        self.capped_raise = CappedRaise()
        if x is not None:
            self.capped_raise.happened_this_round = True
            self.capped_raise.player_that_raised = self.seats[x[0]]
            if x[1] is None:
                self.capped_raise.player_that_cant_reopen = None
            else:
                self.capped_raise.player_that_cant_reopen = self.seats[x[1]]

        self.current_player = self.seats[env_state_dict[EnvDictIdxs.current_player]]

        x = env_state_dict[EnvDictIdxs.last_raiser]
        self.last_raiser = None if x is None else self.seats[x]
        self.deck.load_state_dict(env_state_dict[EnvDictIdxs.deck])
        self.n_actions_this_episode = env_state_dict[EnvDictIdxs.n_actions_this_episode]

        if self.IS_FIXED_LIMIT_GAME:
            self.n_raises_this_round = env_state_dict[EnvDictIdxs.n_raises_this_round]

        for p in self.seats:
            p.seat_id = env_state_dict[EnvDictIdxs.seats][p.seat_id][PlayerDictIdxs.seat_id]
            p.stack = env_state_dict[EnvDictIdxs.seats][p.seat_id][PlayerDictIdxs.stack]
            p.current_bet = env_state_dict[EnvDictIdxs.seats][p.seat_id][PlayerDictIdxs.current_bet]
            p.is_allin = env_state_dict[EnvDictIdxs.seats][p.seat_id][PlayerDictIdxs.is_allin]
            p.folded_this_episode = env_state_dict[EnvDictIdxs.seats][p.seat_id][PlayerDictIdxs.folded_this_episode]
            p.has_acted_this_round = env_state_dict[EnvDictIdxs.seats][p.seat_id][PlayerDictIdxs.has_acted_this_round]
            p.side_pot_rank = env_state_dict[EnvDictIdxs.seats][p.seat_id][PlayerDictIdxs.side_pot_rank]

            if blank_private_info:
                p.hand = None
                p.hand_rank = None
            else:
                p.hand = np.copy(env_state_dict[EnvDictIdxs.seats][p.seat_id][PlayerDictIdxs.hand])
                p.hand_rank = env_state_dict[EnvDictIdxs.seats][p.seat_id][PlayerDictIdxs.hand_rank]

    def get_current_obs(self, is_terminal):
        """
        This function can be useful for manually setting the environment to a desired state and then getting the
        formatted observation from it without actually stepping it.

        Args:
            is_terminal (bool): Whether the last env.step() call returned that it is done or not

        Returns:
            np.ndarray: current observation as returned by the last env.step() call.

        """
        if is_terminal:
            return np.zeros(shape=self.observation_space.shape, dtype=np.float32)  # terminal is all zero
        normalization_sum = float(sum([s.starting_stack_this_episode for s in self.seats])) / self.N_SEATS
        return np.array(self._get_table_state(normalization_sum=normalization_sum) \
                        + self._get_player_states_all_players(normalization_sum=normalization_sum) \
                        + self._get_board_state()
                        , dtype=np.float32)

    def print_obs(self, obs):
        print("______________________________________ Printing _Observation _________________________________________")
        names = [e + ":  " for e in list(self.obs_idx_dict.keys())]
        str_len = max([len(e) for e in names])
        for name, key in zip(names, list(self.obs_idx_dict.keys())):
            name = name.rjust(str_len)
            print(name, obs[self.obs_idx_dict[key]])

    def set_to_public_tree_node_state(self, node):
        """
        Since there might be a conflict between ""node""'s {private cards, remaining deck, board} and those of
        the wrapped env, we need rules:

        - Take the board from the node
        - Private cards from the environment
        - Remaining deck doesn't matter

        This can result in cards appearing multiple times, if misused, but still represents the use-case best.
        """
        self.load_state_dict(node.env_state)

    def cards2str(self, cards_2d, seperator=", "):
        """

        Args:
            cards_2d:           2D representation of any amount of cards
            seperator (str):    token to put between cards when printing

        Returns:
            str

        """
        hand_as_str = ""
        for c in cards_2d:
            if not np.array_equal(c, Poker.CARD_NOT_DEALT_TOKEN_2D):
                hand_as_str += self.RANK_DICT[c[0]]
                hand_as_str += self.SUIT_DICT[c[1]]
                hand_as_str += seperator
        return hand_as_str

    def get_legal_actions(self):
        """

        Returns:
            list:   a list with none, one or multiple actions of the set [FOLD, CHECK/CALL, BET]

        """
        legal_actions = []
        for a in [(Poker.FOLD, -1,), (Poker.CHECK_CALL, -1,)]:
            if self._get_fixed_action(action=a)[0] == a[0]:
                legal_actions.append(a[0])

        a = self._get_env_adjusted_action_formulation(action=(Poker.BET_RAISE, 1,))
        fixed_a = self._get_fixed_action(action=a)
        if a[0] == fixed_a[0]:
            legal_actions.append(Poker.BET_RAISE)

        return legal_actions

    def get_range_idx(self, p_id):
        """
        Args:
            p_id (int):     seat_id of the PokerPlayer whose range_idx (an integer representation of private cards)
                            is requested

        Returns:
            int

        """
        return int(self.lut_holder.get_range_idx_from_hole_cards(
            hole_cards_2d=self.get_hole_cards_of_player(p_id=p_id)))

    def get_random_action(self):
        a = np.random.randint(low=0, high=3)
        pot_sum = sum(self.side_pots) + self.main_pot
        n = int(np.random.normal(loc=pot_sum / 2, scale=pot_sum / 5))

        return a, n

    def get_all_winnable_money(self):
        """

        Returns:
            int: all money in the pot, side pots, and currently bet
        """
        return self.main_pot + sum([p.current_bet for p in self.seats]) + sum(self.side_pots)

    def cards_state_dict(self):
        return {
            "deck": self.deck.state_dict(),
            "board": np.copy(self.board),
            "hand": [np.copy(self.seats[p].hand) for p in range(self.N_SEATS)]
        }

    def load_cards_state_dict(self, cards_state_dict):
        self.deck.load_state_dict(cards_state_dict["deck"])
        self.board = np.copy(cards_state_dict["board"])
        for p in range(self.N_SEATS):
            self.seats[p].hand = np.copy(cards_state_dict["hand"][p])

    def reshuffle_remaining_deck(self):
        self.deck.shuffle()

    def get_fraction_of_pot_raise(self, fraction, player_that_bets):
        """

        Turns a ""fraction_of_pot_raise_size"" into a ""number_of_chips"" raise size.
        We calculate the pot as the sum of all sidepots and the mainpot.

        Args:
            fraction:           fraction of pot to bet
            player_that_bets:   any PokerPlayer in self.seats

        Returns:
            int:                number of chips that produces a raise by the given fraction of the pot
        """
        _player = player_that_bets if isinstance(player_that_bets, PokerPlayer) else self.seats[player_that_bets]
        to_call = self._get_biggest_bet_out_there_aka_total_to_call() - _player.current_bet
        pot_after_call = self.main_pot + sum(self.side_pots) + sum([p.current_bet for p in self.seats]) + to_call

        delta = int(to_call + (pot_after_call * fraction))
        total_raise = delta + _player.current_bet

        return total_raise

    def get_frac_from_chip_amt(self, amt, player_that_bets):
        """
        reverse of self.get_fraction_of_pot_raise()

        Args:
            amt:                number of chips (total bet, not additional!)
            player_that_bets:   any PokerPlayer in self.seats

        Returns:
            float:              fraction of pot that produces a raise by the given amt of chips

        """
        _player = player_that_bets if isinstance(player_that_bets, PokerPlayer) else self.seats[player_that_bets]

        to_call = self._get_biggest_bet_out_there_aka_total_to_call() - _player.current_bet
        pot_after_call = self.main_pot + sum(self.side_pots) + sum([p.current_bet for p in self.seats]) + to_call

        delta = amt - _player.current_bet
        fraction = float(delta - to_call) / float(pot_after_call)

        return fraction

    def get_hole_cards_of_player(self, p_id):
        return self.seats[p_id].hand

    def eval(self):
        """
        Eval mode DOES NOT allow randomizations to be applied.
        """
        self.IS_EVALUATING = True
        for p in self.seats:
            p.IS_EVALUATING = True

    def training(self):
        """
        Allows randomizations as defined in the args passed when creating an env.
        """
        self.IS_EVALUATING = False
        for p in self.seats:
            p.IS_EVALUATING = False

    def render(self, mode='TEXT'):
        """
        When called, renders the current state of the environment

        Args:
            mode (str):   Currently only "TEXT" is supported.

        """

        if mode.upper() == 'TEXT':
            print('\n\n')
            print('___________________________________',
                  (Poker.INT2STRING_ROUND[self.current_round] + " - " + str(
                      self.current_player.seat_id) + " acts").center(15),
                  '___________________________________')

            print("Board: ", self.cards2str(self.board))

            print(("Last Action:   player_" + str(self.last_action[2]) + ": " + str(self.last_action[0]) + " " +
                   str(self.last_action[1]).rjust(113)), "|   Main_pot: ", str(self.main_pot).rjust(7))

            # player information
            for p in self.seats:
                f = "-" if p.folded_this_episode else "+" if p.is_allin else ""
                print((f + "Player_" + str(p.seat_id) + ":").rjust(14) +
                      "stack: ", str(p.stack).rjust(8),
                      "current_bet: ", str(p.current_bet).rjust(8),
                      "side_pot_rank: ", str(p.side_pot_rank).rjust(8),
                      "hand: ", self.cards2str(p.hand).rjust(8),

                      # sidepots
                      " " * 18,
                      '|   Side_pot' + str(p.seat_id) + ": " + str(self.side_pots[p.seat_id]).rjust(6))

            if self.IS_FIXED_LIMIT_GAME:
                print("Num raises this round: ", self.n_raises_this_round)
            print("\n")

        else:
            raise ValueError(mode + "is unsupported")

    def print_tutorial(self):
        print("0 \tFold")
        print("1 \tCall")
        print("2 \tRaise")
        print("If you enter 2 for raise, you will be asked what size you want to raise to TOTAL in case you play a "
              "Poker game with multiple size options")

    def human_api_ask_action(self):
        while True:
            try:
                action_idx = int(
                    input("What action do you want to take as player " + str(self.current_player.seat_id) + "?"))
            except ValueError:
                print("You need to enter one of the allowed actions. Refer to the tutorial please.")
                continue
            if action_idx not in [Poker.FOLD, Poker.CHECK_CALL, Poker.BET_RAISE]:
                print("Invalid action_idx! Please enter one of the allowed actions as described in the tutorial above.")
                continue
            break
        time.sleep(0.01)
        if action_idx == Poker.BET_RAISE:
            raise_size = int(input("To how many chips total do you want to raise?"))
        else:
            raise_size = 0

        action_tuple = [action_idx, raise_size]
        return action_tuple

    def set_stack_size(self, stack_size):
        args = copy.deepcopy(self._args)
        args.starting_stack_sizes_list = copy.deepcopy(stack_size)
        self._init_from_args(env_args=args, is_evaluating=self.IS_EVALUATING)

    def get_args(self):
        return copy.deepcopy(self._args)

    def set_args(self, env_args):
        self._init_from_args(env_args=env_args, is_evaluating=self.IS_EVALUATING)


class CappedRaise:
    """
    There is a special rule in poker games, that says that when player A raises, and player B reraises all-in, but has
    less chips than the minimum raise would actually be, player A cannot rereraise when it is his turn again, while all
    other players can rereraise. This lock is broken when a) someone else rereraises or b) the betting round is over.
    """

    def __init__(self):
        self.happened_this_round = None
        self.player_that_raised = None
        self.player_that_cant_reopen = None
        self.reset()

    def reset(self):
        self.happened_this_round = False
        self.player_that_raised = None
        self.player_that_cant_reopen = None



# Copyright (c) 2019 Eric Steinberger


"""
Wrap a poker environment to track history, for instance. Wrappers are never constructed directly, but only over the
env_builder interface. Creating an env_builder is your starting point to create an environment
(potentially with a wrapper).
"""

from PokerRL.game.Poker import Poker
from PokerRL.game._.EnvWrapperBuilderBase import EnvWrapperBuilderBase as _EnvWrapperBuilderBase
from PokerRL.game._.wrappers.FlatHULimitPokerHistoryWrapper import \
    FlatHULimitPokerHistoryWrapper as _FlatHULimitPokerHistoryWrapper
from PokerRL.game._.wrappers.RecurrentHistoryWrapper import RecurrentHistoryWrapper as _RecurrentHistoryWrapper
from PokerRL.game._.wrappers.Vanilla import VanillaWrapper as _VanillaWrapper


class VanillaEnvBuilder(_EnvWrapperBuilderBase):
    """ For docs please refer to the corresponding wrapper class file. """
    WRAPPER_CLS = _VanillaWrapper


class HistoryEnvBuilder(_EnvWrapperBuilderBase):
    """ For docs please refer to the corresponding wrapper class file. """
    WRAPPER_CLS = _RecurrentHistoryWrapper

    def __init__(self, env_cls, env_args, invert_history_order=False):
        super().__init__(env_cls=env_cls, env_args=env_args)
        self.invert_history_order = invert_history_order


class FlatLimitPokerEnvBuilder(_EnvWrapperBuilderBase):
    """ For docs please refer to the corresponding wrapper class file. """
    WRAPPER_CLS = _FlatHULimitPokerHistoryWrapper

    def __init__(self, env_cls, env_args):
        assert env_cls.IS_FIXED_LIMIT_GAME
        assert env_args.n_seats == 2

        self._VEC_ROUND_OFFSETS = {}
        self._VEC_HALF_ROUND_SIZE = {}
        self.action_vector_size = 0
        for r in env_cls.RULES.ALL_ROUNDS_LIST:
            self._VEC_ROUND_OFFSETS[r] = self.action_vector_size
            self._VEC_HALF_ROUND_SIZE[r] = len([Poker.BET_RAISE, Poker.CHECK_CALL]) * (
                env_cls.MAX_N_RAISES_PER_ROUND[r] + 2)
            self.action_vector_size += self._VEC_HALF_ROUND_SIZE[r] * env_args.n_seats  # always 2 seats

        super().__init__(env_cls=env_cls, env_args=env_args)

    def _get_num_public_observation_features(self):
        _env = self.env_cls(env_args=self.env_args, lut_holder=self.lut_holder, is_evaluating=True)
        return _env.observation_space.shape[0] + self.action_vector_size

    def get_vector_idx(self, round_, p_id, nth_action_this_round, action_idx):
        # *2 in line 3 stands for len([Poker.BET_RAISE, Poker.CHECK_CALL]). If action is fold, the obs is never going to
        # be seen anyways.
        return self._VEC_ROUND_OFFSETS[round_] \
               + p_id * self._VEC_HALF_ROUND_SIZE[round_] \
               + nth_action_this_round * 2 \
               + action_idx - 1  # - 1 because fold (which is action_idx==0) is not recorded.


ALL_BUILDERS = [
    HistoryEnvBuilder,
    FlatLimitPokerEnvBuilder,
    VanillaEnvBuilder
]



# Copyright (c) 2019 Eric Steinberger


import numpy as np

from PokerRL.game.Poker import Poker


class PokerRange:
    """
    Track a distribution over hands for a player.
    """

    def __init__(self, env_bldr):
        assert env_bldr.rules.N_HOLE_CARDS <= 2

        self._env_bldr = env_bldr
        self._range = None  # np.array(range_size)
        self.reset()

    @property
    def range(self):
        return self._range

    def get_range(self):
        return np.copy(self._range)

    def get_card_probs(self):
        if self._env_bldr.rules.N_HOLE_CARDS == 1:
            return np.copy(self._range)
        elif self._env_bldr.rules.N_HOLE_CARDS == 2:
            card_probs = np.zeros(shape=self._env_bldr.rules.N_CARDS_IN_DECK, dtype=np.float32)

            for c in range(self._env_bldr.rules.N_CARDS_IN_DECK):
                card_probs[c] = np.sum(self._range[self._env_bldr.lut_holder.LUT_CARD_IN_WHAT_RANGE_IDXS[c]])
            return card_probs

        else:
            raise NotImplementedError()

    def mul_and_norm(self, mul_vector):
        self._range *= mul_vector
        self.normalize()

    def normalize(self):
        norm_factor = np.sum(self._range, axis=-1)
        if norm_factor == 0:
            self._reset_range()
        else:
            self._range = self._range / norm_factor

    def update_after_action(self, action, all_a_probs_for_all_hands):
        self._range *= all_a_probs_for_all_hands[:, action]
        self.normalize()

    def update_after_new_round(self, new_round, board_now_2d):
        """
        1) Remove the new blockers from the range by setting all hands including them to probability 0
        2) Normalize
        """
        self.set_cards_to_zero_prob(cards_2d=self._get_new_blockers_2d(game_round=new_round, board_2d=board_now_2d))

    def reset(self):
        """
        Before any cards are dealt ranges are uniformly random distributions
        """
        self._reset_range()

    def set_cards_to_zero_prob(self, cards_2d):
        cards_1d_to_remove = self._env_bldr.lut_holder.get_1d_cards(cards_2d=cards_2d)
        if self._env_bldr.rules.N_HOLE_CARDS == 1:
            self._range[cards_1d_to_remove] = 0

        elif self._env_bldr.rules.N_HOLE_CARDS == 2:
            for c in cards_1d_to_remove:
                # instead of looping we make use of the fact that the LUT is sorted and just use numpy slicing
                self._range[self._env_bldr.lut_holder.LUT_HOLE_CARDS_2_IDX[0:c, c]] = 0
                self._range[
                    self._env_bldr.lut_holder.LUT_HOLE_CARDS_2_IDX[c, c + 1:self._env_bldr.rules.N_CARDS_IN_DECK]] = 0
        else:
            raise NotImplementedError("We don't currently support games with >2 hole cards")

        self.normalize()

    @staticmethod
    def get_possible_range_idxs(rules, lut_holder, board_2d):
        arr = np.arange(rules.RANGE_SIZE)

        if board_2d.shape[0] == 0:  # if nothing is blocked
            return arr

        # filter not-dealt-cards
        lut_holder.get_1d_cards(cards_2d=board_2d)
        blocked_cards_1d = np.array(
            [c for c in lut_holder.get_1d_cards(cards_2d=board_2d) if c != Poker.CARD_NOT_DEALT_TOKEN_1D])

        if rules.N_HOLE_CARDS == 1:
            arr = np.delete(arr, obj=blocked_cards_1d)
            return arr

        elif rules.N_HOLE_CARDS == 2:
            hands = []

            for c in blocked_cards_1d:
                for c1 in range(0, c):
                    hands.append(lut_holder.get_range_idx_from_hole_cards(
                        lut_holder.get_2d_cards(np.array([c1, c], dtype=np.int8))))
                for c2 in range(c + 1, rules.N_CARDS_IN_DECK):
                    hands.append(lut_holder.get_range_idx_from_hole_cards(
                        lut_holder.get_2d_cards(np.array([c, c2], dtype=np.int8))))

            blocked_idxs = np.unique(np.array(hands))
            arr = np.delete(arr, obj=blocked_idxs)

            return arr

        else:
            raise NotImplementedError("self.N_HOLE_CARDS > 2:  " + str(rules.N_HOLE_CARDS))

    @staticmethod
    def get_range_size(n_hole_cards, n_cards_in_deck):
        """
        Args:
            n_hole_cards:       number of cards each player is dealt
            n_cards_in_deck:    number of unique cards in the deck

        Returns:
            int:                the number of possible hands (order of cards does not matter) given a set number of
                                holecards and cards in the deck.
        """
        range_size = 1
        for i in range(n_hole_cards):
            range_size *= n_cards_in_deck - i

        n_hc_factorial = np.prod(np.arange(1, n_hole_cards + 1))
        return int(range_size / n_hc_factorial)

    def load_state_dict(self, state):
        self._range = np.copy(state['range'])

    def state_dict(self):
        return {
            'range': np.copy(self._range),
        }

    def _get_new_blockers_1d(self, game_round, board_2d):
        _1d = self._env_bldr.lut_holder.get_1d_cards(
            self._get_new_blockers_2d(game_round=game_round, board_2d=board_2d))
        return _1d

    def _get_new_blockers_2d(self, game_round, board_2d):
        n = self._env_bldr.lut_holder.DICT_LUT_N_CARDS_OUT[game_round]
        nm1 = self._env_bldr.lut_holder.DICT_LUT_N_CARDS_OUT[self._env_bldr.rules.ROUND_BEFORE[game_round]]
        new_blockers = board_2d[nm1:n].reshape(-1, 2)
        return new_blockers

    def _reset_range(self):
        self._range = np.full(shape=self._env_bldr.rules.RANGE_SIZE,
                              fill_value=1.0 / self._env_bldr.rules.RANGE_SIZE,
                              dtype=np.float32)



# Copyright (c) 2019 Eric Steinberger


import ray

from PokerRL.eval.rl_br.workers.ps.Local_RLBR_ParameterServer import Local_RLBR_ParameterServer as _LocalPS


@ray.remote(num_cpus=1)
class Dist_RLBR_ParameterServer(_LocalPS):

    def __init__(self, t_prof, chief_handle):
        super().__init__(t_prof=t_prof, chief_handle=chief_handle)



# Copyright (c) 2019 Eric Steinberger


from PokerRL.game._.rl_env.base.PokerEnv import PokerEnv as _PokerEnv

from PokerRL.game.poker_env_args import NoLimitPokerEnvArgs


class NoLimitPokerEnv(_PokerEnv):
    """
    Since the PokerEnv base-class is written to be no-limit by default, this wrapper just passes straight through; it
    only exists for consistent style.
    """
    ARGS_CLS = NoLimitPokerEnvArgs

    def __init__(self,
                 env_args,
                 lut_holder,
                 is_evaluating):
        """
        Args:
            env_args (DiscretePokerEnvArgs):    an instance of DiscretePokerEnvArgs, passing an instance of PokerEnvArgs
                                                will not work.
            is_evaluating (bool):               Whether the environment shall be spawned in evaluation mode
                                                (i.e. no randomization) or not.

        """
        assert isinstance(env_args, NoLimitPokerEnvArgs)
        super().__init__(env_args=env_args, lut_holder=lut_holder, is_evaluating=is_evaluating)
        self.N_ACTIONS = env_args.N_ACTIONS

    def _adjust_raise(self, raise_total_amount_in_chips):
        return max(self._get_current_total_min_raise(), raise_total_amount_in_chips)



# Copyright (c) 2019 Eric Steinberger


import torch.nn as nn


class AdvantageNet(nn.Module):

    def __init__(self, env_bldr, adv_net_args, device):
        super().__init__()

        self._env_bldr = env_bldr
        self._args = adv_net_args
        self._n_actions = env_bldr.N_ACTIONS

        self._relu = nn.ReLU(inplace=False)

        MPM = adv_net_args.mpm_args.get_mpm_cls()
        self._mpm = MPM(env_bldr=env_bldr, device=device, mpm_args=adv_net_args.mpm_args)

        self._final_layer = nn.Linear(in_features=self._mpm.output_units,
                                      out_features=adv_net_args.n_units_final)
        self._adv = nn.Linear(in_features=adv_net_args.n_units_final, out_features=self._n_actions)

        self.to(device)

    def forward(self, pub_obses, range_idxs, legal_action_masks):
        y = self._mpm(pub_obses=pub_obses, range_idxs=range_idxs)
        y = self._relu(self._final_layer(y))
        y = self._adv(y)

        # sets all illegal actions to 0 for mean computation
        y *= legal_action_masks

        # can't directly compute mean cause illegal actions are still in there. Computing sum and dividing by ||A(I)||
        mean = (y.sum(dim=1) / legal_action_masks.sum(dim=1)).unsqueeze(1).expand(-1, self._n_actions)

        # subtracting mean also subtracts from illegal actions; have to mask again
        return (y - mean) * legal_action_masks


class AdvNetArgs:

    def __init__(self, n_units_final, mpm_args):
        self.n_units_final = n_units_final
        self.mpm_args = mpm_args



# Copyright (c) 2019 Eric Steinberger


import copy

import numpy as np

from PokerRL.game._.tree.PublicTree import PublicTree
from PokerRL.game.wrappers import HistoryEnvBuilder
from PokerRL.rl.rl_util import get_env_cls_from_str


class CFRBase:
    """
    base class to all full-width (i.e. non MC) tabular CFR methods
    """

    def __init__(self,
                 name,
                 chief_handle,
                 game_cls,
                 agent_bet_set,
                 algo_name,
                 starting_stack_sizes=None,
                 ):
        """
        Args:
            name (str):                             Under this name all logs, data, and checkpoints will appear.
            chief_handle (ChiefBase):               Reference to chief worker
            game_cls (PokerEnv subclass):           Class (not instance) to be trained in.
            agent_bet_set (iterable):               Choosing a bet-set from bet_sets.py is recommended. If solving a
                                                    Limit poker game, this value will not be considered, but must still
                                                    be passed. Just set this to any list of floats (e.g. [0.0])
            starting_stack_sizes (list of ints):    For each stack size in this list, a CFR strategy will be computed.
                                                    Results are logged individually and averaged (uniform).
                                                    If None, takes the default for the game.
        """

        self._name = name
        self._n_seats = 2

        self._chief_handle = chief_handle

        if starting_stack_sizes is None:
            self._starting_stack_sizes = [game_cls.DEFAULT_STACK_SIZE]
        else:
            self._starting_stack_sizes = copy.deepcopy(starting_stack_sizes)
        self._game_cls_str = game_cls.__name__

        self._env_args = [
            game_cls.ARGS_CLS(n_seats=self._n_seats,
                              starting_stack_sizes_list=[start_chips for _ in range(self._n_seats)],
                              bet_sizes_list_as_frac_of_pot=agent_bet_set,
                              )
            for start_chips in self._starting_stack_sizes
        ]
        self._env_bldrs = [
            HistoryEnvBuilder(env_cls=get_env_cls_from_str(self._game_cls_str),
                              env_args=self._env_args[s])

            for s in range(len(self._starting_stack_sizes))
        ]

        self._trees = [
            PublicTree(env_bldr=self._env_bldrs[idx],
                       stack_size=self._env_args[idx].starting_stack_sizes_list,
                       stop_at_street=None)
            for idx in range(len(self._env_bldrs))
        ]

        for tree in self._trees:
            tree.build_tree()
            print("Tree with stack size", tree.stack_size, "has", tree.n_nodes, "nodes out of which", tree.n_nonterm,
                  "are non-terminal.")

        self._algo_name = algo_name

        self._exps_curr_total = [
            self._chief_handle.create_experiment(
                self._name + "_Curr_S" + str(self._starting_stack_sizes[s]) + "_total_" + self._algo_name)
            for s in range(len(self._starting_stack_sizes))
        ]

        self._exps_avg_total = [
            self._chief_handle.create_experiment(
                self._name + "_Avg_total_S" + str(self._starting_stack_sizes[s]) + "_" + self._algo_name)
            for s in range(len(self._starting_stack_sizes))
        ]

        self._exp_all_averaged_curr_total = self._chief_handle.create_experiment(
            self._name + "_Curr_total_averaged_" + self._algo_name)

        self._exp_all_averaged_avg_total = self._chief_handle.create_experiment(
            self._name + "_Avg_total_averaged_" + self._algo_name)

        self._iter_counter = None

    @property
    def name(self):
        return self._name

    @property
    def algo_name(self):
        return self._algo_name

    @property
    def iter_counter(self):
        return self._iter_counter

    def reset(self):
        self._iter_counter = 0

        for p in range(self._n_seats):
            self._reset_player(p_id=p)

        for t_idx in range(len(self._trees)):
            self._trees[t_idx].fill_uniform_random()

        self._compute_cfv()
        self._log_curr_strat_expl()

    def iteration(self):
        for p in range(self._n_seats):
            self._compute_cfv()
            self._compute_regrets(p_id=p)
            self._compute_new_strategy(p_id=p)
            self._update_reach_probs()
            self._add_strategy_to_average(p_id=p)

        self._iter_counter += 1

        self._compute_cfv()
        self._log_curr_strat_expl()
        self._evaluate_avg_strats()

    def _compute_cfv(self):
        for t_idx in range(len(self._trees)):
            self._trees[t_idx].compute_ev()

    def _regret_formula_first_it(self, ev_all_actions, strat_ev):
        raise NotImplementedError

    def _regret_formula_after_first_it(self, ev_all_actions, strat_ev, last_regrets):
        raise NotImplementedError

    def _compute_regrets(self, p_id):

        for t_idx in range(len(self._trees)):
            def __compute_evs(_node):
                # EV of each action
                N_ACTIONS = len(_node.children)
                ev_all_actions = np.zeros(shape=(self._env_bldrs[t_idx].rules.RANGE_SIZE, N_ACTIONS), dtype=np.float32)
                for i, child in enumerate(_node.children):
                    ev_all_actions[:, i] = child.ev[p_id]

                # EV if playing by curr strat
                strat_ev = _node.ev[p_id]
                strat_ev = np.expand_dims(strat_ev, axis=-1).repeat(N_ACTIONS, axis=-1)

                return strat_ev, ev_all_actions

            def _fill_after_first(_node):
                if _node.p_id_acting_next == p_id:
                    strat_ev, ev_all_actions = __compute_evs(_node=_node)
                    _node.data["regret"] = self._regret_formula_after_first_it(ev_all_actions=ev_all_actions,
                                                                               strat_ev=strat_ev,
                                                                               last_regrets=_node.data["regret"])

                for c in _node.children:
                    _fill_after_first(c)

            def _fill_first(_node):
                if _node.p_id_acting_next == p_id:
                    strat_ev, ev_all_actions = __compute_evs(_node=_node)

                    _node.data["regret"] = self._regret_formula_first_it(ev_all_actions=ev_all_actions,
                                                                         strat_ev=strat_ev)

                for c in _node.children:
                    _fill_first(c)

            if self._iter_counter == 0:
                _fill_first(self._trees[t_idx].root)
            else:
                _fill_after_first(self._trees[t_idx].root)

    def _compute_new_strategy(self, p_id):
        """ Assumes regrets have been computed for player ""p_id"" already! """
        raise NotImplementedError

    def _update_reach_probs(self):
        for t_idx in range(len(self._trees)):
            self._trees[t_idx].update_reach_probs()

    def _add_strategy_to_average(self, p_id):
        raise NotImplementedError

    def _log_curr_strat_expl(self):
        expl_totals = []
        for t_idx in range(len(self._trees)):
            METRIC = self._env_bldrs[t_idx].env_cls.WIN_METRIC
            expl_p = [
                float(self._trees[t_idx].root.exploitability[p]) * self._env_bldrs[t_idx].env_cls.EV_NORMALIZER
                for p in range(self._n_seats)
            ]
            expl_total = sum(expl_p) / self._n_seats
            expl_totals.append(expl_total)

            self._chief_handle.add_scalar(self._exps_curr_total[t_idx],
                                          "Evaluation/" + METRIC, self._iter_counter, expl_total)

            self._trees[t_idx].export_to_file(name=self._name + "_Curr_" + str(self._iter_counter))

        expl_total_averaged = sum(expl_totals) / float(len(expl_totals))
        self._chief_handle.add_scalar(self._exp_all_averaged_curr_total,
                                      "Evaluation/" + METRIC, self._iter_counter, expl_total_averaged)

    def _evaluate_avg_strats(self):
        expl_totals = []
        for t_idx in range(len(self._trees)):
            METRIC = self._env_bldrs[t_idx].env_cls.WIN_METRIC
            eval_tree = PublicTree(env_bldr=self._env_bldrs[t_idx],
                                   stack_size=self._env_args[t_idx].starting_stack_sizes_list,
                                   stop_at_street=None,
                                   is_debugging=False,
                                   )
            eval_tree.build_tree()

            def _fill(_node_eval, _node_train):
                if _node_eval.p_id_acting_next != eval_tree.CHANCE_ID and (not _node_eval.is_terminal):
                    _node_eval.strategy = np.copy(_node_train.data["avg_strat"])
                    assert np.allclose(np.sum(_node_eval.strategy, axis=1), 1, atol=0.0001)

                for c_eval, c_train in zip(_node_eval.children, _node_train.children):
                    _fill(_node_eval=c_eval, _node_train=c_train)

            # sets up some stuff; we overwrite strategy afterwards
            eval_tree.fill_uniform_random()

            # fill with strat
            _fill(_node_eval=eval_tree.root, _node_train=self._trees[t_idx].root)
            eval_tree.update_reach_probs()

            # compute EVs
            eval_tree.compute_ev()

            eval_tree.export_to_file(name=self._name + "_Avg_" + str(self._iter_counter))

            # log
            expl_p = [
                float(eval_tree.root.exploitability[p]) * self._env_bldrs[t_idx].env_cls.EV_NORMALIZER
                for p in range(eval_tree.n_seats)
            ]
            expl_total = sum(expl_p) / eval_tree.n_seats
            expl_totals.append(expl_total)

            self._chief_handle.add_scalar(self._exps_avg_total[t_idx],
                                          "Evaluation/" + METRIC, self._iter_counter, expl_total)

        expl_total_averaged = sum(expl_totals) / float(len(expl_totals))
        self._chief_handle.add_scalar(self._exp_all_averaged_avg_total,
                                      "Evaluation/" + METRIC, self._iter_counter, expl_total_averaged)

    def _reset_player(self, p_id):
        def __reset(_node, _p_id):
            if _node.p_id_acting_next == _p_id:
                # regrets and strategies only need to be stored for one player at each node
                _node.data = {
                    "regret": None,
                    "avg_strat": None
                }
                _node.strategy = None

            for c in _node.children:
                __reset(c, _p_id=_p_id)

        for t_idx in range(len(self._trees)):
            __reset(self._trees[t_idx].root, _p_id=p_id)



# Copyright (c) 2019 Eric Steinberger


"""
The local master contains logic to manage many workers, but if instantiated as local, will only manage one. It handles
logging and provides an interface to the LBR computation.
"""
import numpy as np

from PokerRL.eval._.EvaluatorMasterBase import EvaluatorMasterBase
from PokerRL.eval.lbr import _util


class LocalLBRMaster(EvaluatorMasterBase):
    """
    LBR computation as described in https://arxiv.org/abs/1612.07547
    
    EvalLBRMaster manages a sub-cluster of EvalLBRWorker nodes.
    """

    def __init__(self, t_prof, chief_handle):
        assert t_prof.n_seats == 2

        EvaluatorMasterBase.__init__(self, t_prof=t_prof, eval_env_bldr=_util.get_env_builder_lbr(t_prof=t_prof),
                                     chief_handle=chief_handle, eval_type="LBR", log_conf_interval=True)

        self.lbr_args = t_prof.module_args["lbr"]

        self.weights_for_eval_agent = None
        self.alive_worker_handles = None

    def set_worker_handles(self, *worker_handles):
        self.alive_worker_handles = list(worker_handles)

    def evaluate(self, iter_nr):
        # __________________________________ send weights from Master to all Workers ___________________________________
        self._ray.wait([
            self._ray.remote(
                worker.update_weights,
                self.weights_for_eval_agent
            )
            for worker in self.alive_worker_handles
        ])

        # _______________________________ Evaluate for all specified eval_modes_of_algo ________________________________
        for mode in self._t_prof.eval_modes_of_algo:
            if self._is_multi_stack:
                total_of_all_stacks = []
                pm_of_all_stacks = []

            for stack_size_idx, stack_size in enumerate(self._t_prof.eval_stack_sizes):
                scores = []
                for p_id in range(self._t_prof.n_seats):
                    scores += (self._ray.get(
                        [
                            self._ray.remote(worker.run,
                                             p_id,
                                             int(self.lbr_args.n_lbr_hands / self.lbr_args.n_workers),
                                             mode,
                                             stack_size
                                             )
                            for worker in self.alive_worker_handles
                        ]
                    ))

                scores = [s for s in scores if s is not None]
                scores = np.concatenate(scores, axis=0)
                if len(scores) > 0:
                    mean, d = self._get_95confidence(scores)

                    self._log_results(iter_nr=iter_nr,
                                      agent_mode=mode,
                                      stack_size_idx=stack_size_idx,
                                      score=mean, upper_conf95=mean + d, lower_conf95=mean - d)

                    if self._is_multi_stack:
                        total_of_all_stacks.append(mean)

            if self._is_multi_stack:
                _mean = sum(total_of_all_stacks) / float(len(total_of_all_stacks))
                _d = sum(pm_of_all_stacks) / float(len(pm_of_all_stacks))
                self._log_multi_stack(agent_mode=mode,
                                      iter_nr=iter_nr,
                                      score_total=_mean,
                                      upper_conf95=_mean + _d,
                                      lower_conf95=_mean - _d,
                                      )

    def update_weights(self):
        self.weights_for_eval_agent = self.pull_current_strat_from_chief()



# Copyright (c) 2019 Eric Steinberger


import numpy as np

from PokerRL.cfr._CFRBase import CFRBase as _CFRBase


class VanillaCFR(_CFRBase):

    def __init__(self,
                 name,
                 chief_handle,
                 game_cls,
                 agent_bet_set,
                 starting_stack_sizes=None):
        super().__init__(name=name,
                         chief_handle=chief_handle,
                         game_cls=game_cls,
                         starting_stack_sizes=starting_stack_sizes,
                         agent_bet_set=agent_bet_set,
                         algo_name="CFR"
                         )
        self.reset()

    def _regret_formula_after_first_it(self, ev_all_actions, strat_ev, last_regrets):
        return ev_all_actions - strat_ev + last_regrets

    def _regret_formula_first_it(self, ev_all_actions, strat_ev):
        return ev_all_actions - strat_ev

    def _compute_new_strategy(self, p_id):
        for t_idx in range(len(self._trees)):
            def _fill(_node):
                if _node.p_id_acting_next == p_id:
                    N = len(_node.children)

                    _capped_reg = np.maximum(_node.data["regret"], 0)
                    _reg_pos_sum = np.expand_dims(np.sum(_capped_reg, axis=1), axis=1).repeat(N, axis=1)

                    with np.errstate(divide='ignore', invalid='ignore'):
                        _node.strategy = np.where(
                            _reg_pos_sum > 0.0,
                            _capped_reg / _reg_pos_sum,
                            np.full(shape=(self._env_bldrs[t_idx].rules.RANGE_SIZE, N,), fill_value=1.0 / N,
                                    dtype=np.float32)
                        )

                for c in _node.children:
                    _fill(c)

            _fill(self._trees[t_idx].root)

    def _add_strategy_to_average(self, p_id):
        def _fill(_node):
            if _node.p_id_acting_next == p_id:
                contrib = _node.strategy * np.expand_dims(_node.reach_probs[p_id], axis=1)
                if self._iter_counter > 0:
                    _node.data["avg_strat_sum"] += contrib
                else:
                    _node.data["avg_strat_sum"] = contrib

                _s = np.expand_dims(np.sum(_node.data["avg_strat_sum"], axis=1), axis=1)

                with np.errstate(divide='ignore', invalid='ignore'):
                    _node.data["avg_strat"] = np.where(_s == 0,
                                                       np.full(shape=len(_node.allowed_actions),
                                                               fill_value=1.0 / len(_node.allowed_actions)),
                                                       _node.data["avg_strat_sum"] / _s
                                                       )
                assert np.allclose(np.sum(_node.data["avg_strat"], axis=1), 1, atol=0.0001)

            for c in _node.children:
                _fill(c)

        for t_idx in range(len(self._trees)):
            _fill(self._trees[t_idx].root)



# Copyright (c) 2019 Eric Steinberger


from os.path import join as ospj

from pycrayon import CrayonClient

from PokerRL.rl.MaybeRay import MaybeRay
from PokerRL.util.file_util import create_dir_if_not_exist, write_dict_to_file_json


class CrayonWrapper:
    """
    Wraps PyCrayon (https://github.com/torrvision/crayon), a language-agnostic interface to TensorBoard.
    """

    def __init__(self, name, runs_distributed, runs_cluster, chief_handle, path_log_storage=None,
                 crayon_server_address="localhost"):
        self._name = name
        self._path_log_storage = path_log_storage
        if path_log_storage is not None:
            create_dir_if_not_exist(path_log_storage)

        self._chief_handle = chief_handle
        self._crayon = CrayonClient(hostname=crayon_server_address)
        self._experiments = {}
        self.clear()
        self._custom_logs = {}  # dict of exps containing dict of graph names containing lists of {step: val, } dicts

        self._ray = MaybeRay(runs_distributed=runs_distributed, runs_cluster=runs_cluster)

    @property
    def name(self):
        return self._name

    @property
    def path_log_storage(self):
        return self._path_log_storage

    def clear(self):
        """
        Does NOT clear crayon's internal experiment logs and files.
        """
        self._experiments = {}

    def export_all(self, iter_nr):
        """
        Exports all logs of the current run in Tensorboard's format and as json files.
        """
        if self._path_log_storage is not None:
            path_crayon = ospj(self._path_log_storage, str(self._name), str(iter_nr), "crayon")
            path_json = ospj(self._path_log_storage, str(self._name), str(iter_nr), "as_json")
            create_dir_if_not_exist(path=path_crayon)
            create_dir_if_not_exist(path=path_json)
            for e in self._experiments.values():
                e.to_zip(filename=ospj(path_crayon, e.xp_name + ".zip"))
                write_dict_to_file_json(dictionary=self._custom_logs, _dir=path_json, file_name="logs")

    def update_from_log_buffer(self):
        """
        Pulls newly added logs from the chief onto whatever worker CrayonWrapper runs on. It then adds all these new
        logs to Tensorboard (i.e. PyCrayon's docker container)
        """
        new_v, exp_names = self._get_new_vals()

        for e in exp_names:
            if e not in self._experiments.keys():
                self._custom_logs[e] = {}
                try:
                    self._experiments[e] = self._crayon.create_experiment(xp_name=e)
                except ValueError:
                    self._crayon.remove_experiment(xp_name=e)
                    self._experiments[e] = self._crayon.create_experiment(xp_name=e)

        for name, vals_dict in new_v.items():
            for graph_name, data_points in vals_dict.items():

                for data_point in data_points:
                    step = int(data_point[0])
                    val = data_point[1]

                    self._experiments[name].add_scalar_value(name=graph_name, step=step, value=val)
                    if graph_name not in self._custom_logs[name].keys():
                        self._custom_logs[name][graph_name] = []

                    self._custom_logs[name][graph_name].append({step: val})

    def _get_new_vals(self):
        """
        Returns:
            dict: Pulls and returns newly added logs from the chief onto whatever worker CrayonWrapper runs on.
        """
        return self._ray.get(self._ray.remote(self._chief_handle.get_new_values))



# Copyright (c) 2019 Eric Steinberger


"""
Utility functions for RL
"""
import numpy as np
import torch

from PokerRL.game.games import ALL_ENVS
from PokerRL.game.wrappers import ALL_BUILDERS


def polynomial_decay(base, const, counter, exponent, minimum=0):
    return minimum + ((base - minimum) / (1 + const * np.power(counter, exponent)))


def str_to_optim_cls(optim_string):
    if optim_string.lower() == "sgd":
        return torch.optim.SGD

    elif optim_string.lower() == "adam":
        def fn(parameters, lr):
            return torch.optim.Adam(parameters, lr=lr)

        return fn

    elif optim_string.lower() == "rms":
        def fn(parameters, lr):
            return torch.optim.RMSprop(parameters, lr=lr)

        return fn

    elif optim_string.lower() == "sgdmom":
        def fn(parameters, lr):
            return torch.optim.SGD(parameters, lr=lr, momentum=0.9, nesterov=True)

        return fn

    else:
        raise ValueError(optim_string)


def str_to_loss_cls(loss_str):
    if loss_str.lower() == "mse":
        return torch.nn.MSELoss()

    elif loss_str.lower() == "weighted_mse":
        return lambda y, trgt, w: torch.mean(w * ((y - trgt) ** 2))

    elif loss_str.lower() == "ce":
        return torch.nn.CrossEntropyLoss()

    elif loss_str.lower() == "smoothl1":
        return torch.nn.SmoothL1Loss()

    else:
        raise ValueError(loss_str)


def str_to_rnn_cls(rnn_str):
    if rnn_str.lower() == "lstm":
        return torch.nn.LSTM

    elif rnn_str.lower() == "gru":
        return torch.nn.GRU

    elif rnn_str.lower() == "vanilla":
        return torch.nn.RNN

    else:
        raise ValueError(rnn_str)


def get_env_cls_from_str(env_str):
    for e in ALL_ENVS:
        if env_str == e.__name__:
            return e
    raise ValueError(env_str, "is not registered or does not exist.")


def get_env_builder(t_prof):
    ENV_BUILDER = get_builder_from_str(t_prof.env_builder_cls_str)
    return ENV_BUILDER(env_cls=get_env_cls_from_str(t_prof.game_cls_str), env_args=t_prof.module_args["env"])


def get_builder_from_str(wrapper_str):
    for b in ALL_BUILDERS:
        if wrapper_str == b.__name__:
            return b
    raise ValueError(wrapper_str, "is not registered or does not exist.")


def get_legal_action_mask_torch(n_actions, legal_actions_list, device, dtype=torch.uint8):
    """
    Args:
        legal_actions_list (list):  List of legal actions as integers, where 0 is always FOLD, 1 is CHECK/CALL.
                                    2 is BET/RAISE for continuous PokerEnvs, and for DiscretePokerEnv subclasses,
                                    numbers greater than 1 are all the raise sizes.

        device (torch.device):      device the mask shall be put on

        dtype:                      dtype the mask shall have

    Returns:
        torch.Tensor:               a many-hot representation of the list of legal actions.
    """
    idxs = torch.LongTensor(legal_actions_list, device=device)
    mask = torch.zeros((n_actions), device=device, dtype=dtype)
    mask[idxs] = 1
    return mask


def batch_get_legal_action_mask_torch(n_actions, legal_actions_lists, device, dtype=torch.uint8):
    """

    Args:
        legal_actions_lists (list): List of lists. Each of the 2nd level lists contains legal actions as integers,
                                    where 0 is always FOLD, 1 is CHECK/CALL. 2 is BET/RAISE for continuous
                                    PokerEnvs, and for DiscretePokerEnv subclasses, numbers greater than 1 are all
                                    the raise sizes.

        device (torch.device):      device the mask shall be put on

        dtype:                      dtype the mask shall have

    Returns:
        torch.Tensor:               a many-hot representation of the list of legal actions.

    """
    assert isinstance(legal_actions_lists[0], list), "need list of lists of legal actions (as ints)!"

    mask = torch.zeros((len(legal_actions_lists), n_actions,), device=device, dtype=dtype)
    for i, legal_action_list in enumerate(legal_actions_lists):
        mask[i, torch.LongTensor(legal_action_list, device=device)] = 1
    return mask


def get_legal_action_mask_np(n_actions, legal_actions_list, dtype=np.uint8):
    """

    Args:
        legal_actions_list (list):  List of legal actions as integers, where 0 is always FOLD, 1 is CHECK/CALL.
                                    2 is BET/RAISE for continuous PokerEnvs, and for DiscretePokerEnv subclasses,
                                    numbers greater than 1 are all the raise sizes.

        dtype:                      dtype the mask shall have

    Returns:
        np.ndarray:                 a many-hot representation of the list of legal actions.

    """
    mask = np.zeros(shape=n_actions, dtype=dtype)
    mask[legal_actions_list] = 1
    return mask



# Copyright (c) 2019 Eric Steinberger


import copy
import unittest
from unittest import TestCase

import numpy as np

from PokerRL.game.games import NoLimitHoldem


class TestPokerEnv(TestCase):
    ITERATIONS = 30
    MIN_P = 2
    MAX_P = 6
    DO_RENDER = False

    def test_sync_deck(self):
        args_1 = NoLimitHoldem.ARGS_CLS(n_seats=3,
                                        stack_randomization_range=(0, 0),
                                        starting_stack_sizes_list=[20000] * 3)
        args_2 = NoLimitHoldem.ARGS_CLS(n_seats=3,
                                        stack_randomization_range=(-12, 34),
                                        starting_stack_sizes_list=[123] * 3)
        env1 = NoLimitHoldem(is_evaluating=False, env_args=args_1, lut_holder=NoLimitHoldem.get_lut_holder())
        env2 = NoLimitHoldem(is_evaluating=False, env_args=args_2, lut_holder=NoLimitHoldem.get_lut_holder())

        env1.load_cards_state_dict(cards_state_dict=env2.cards_state_dict())

        for n in range(env2.N_CARDS_IN_DECK - env2.N_HOLE_CARDS * env2.N_SEATS):
            assert np.array_equal(env1.deck.draw(1), env2.deck.draw(1))

        env1.reset()
        env2.reset()
        env1.load_cards_state_dict(cards_state_dict=env2.cards_state_dict())

        for n in range(int((env2.N_CARDS_IN_DECK - (env2.N_HOLE_CARDS * env2.N_SEATS)) / 2)):
            assert np.array_equal(env1.deck.draw(2), env2.deck.draw(2))

    def test_get_current_obs(self):
        args = NoLimitHoldem.ARGS_CLS(n_seats=3,
                                      stack_randomization_range=(0, 0),
                                      starting_stack_sizes_list=[1000] * 3)
        env = NoLimitHoldem(is_evaluating=False, env_args=args, lut_holder=NoLimitHoldem.get_lut_holder())
        env.reset()
        env.step([1, -1])
        a = env.get_current_obs(is_terminal=False)
        b = env.get_current_obs(is_terminal=False)
        assert np.array_equal(a, b)

        # terminal should be all 0
        assert np.array_equal(np.zeros_like(a), env.get_current_obs(is_terminal=True))

    def test_seat_id(self):
        for n in range(TestPokerEnv.MIN_P, TestPokerEnv.MAX_P + 1):
            env = _get_new_nlh_env(n)
            for p_idx in range(len(env.seats)):
                assert env.seats[p_idx].seat_id == p_idx

    def test_chip_consistency_all_stacks_equal(self):
        for n in range(TestPokerEnv.MIN_P, TestPokerEnv.MAX_P + 1):

            for _ in range(TestPokerEnv.ITERATIONS):

                env = _get_new_nlh_env(n)
                original_stack_sum = sum([p.stack for p in env.seats])

                terminal = False
                env.reset()
                if TestPokerEnv.DO_RENDER:
                    env.render()

                while not terminal:
                    obs, reward, terminal, info = env.step(action=env.get_random_action())

                    if TestPokerEnv.DO_RENDER:
                        env.render()

                current_stack_sum = sum([p.stack for p in env.seats])
                assert current_stack_sum == original_stack_sum

    def test_chip_consistency_random_stacks(self):
        for n in range(TestPokerEnv.MIN_P, TestPokerEnv.MAX_P + 1):

            for _ in range(TestPokerEnv.ITERATIONS):

                env = _get_new_nlh_env(n, 100, 1000, True)
                original_stack_sum = sum([p.stack for p in env.seats])

                terminal = False
                env.reset()
                if TestPokerEnv.DO_RENDER:
                    env.render()

                while not terminal:
                    obs, reward, terminal, info = env.step(action=env.get_random_action())
                    if TestPokerEnv.DO_RENDER:
                        env.render()

                current_stack_sum = sum([p.stack for p in env.seats])
                assert current_stack_sum == original_stack_sum

    def test_reward_is_0_for_not_terminal(self):
        for n in range(TestPokerEnv.MIN_P, TestPokerEnv.MAX_P + 1):
            env = _get_new_nlh_env(n, 100, 1000)
            for _ in range(TestPokerEnv.ITERATIONS):

                terminal = False
                env.reset()
                if TestPokerEnv.DO_RENDER:
                    env.render()
                while not terminal:
                    obs, reward, terminal, info = env.step(action=env.get_random_action())
                    if TestPokerEnv.DO_RENDER:
                        env.render()
                    if not terminal:
                        assert np.array_equal(reward, np.zeros(shape=env.N_SEATS, dtype=np.int32))

                env.reset()

    def test_action_space_sample(self):
        env = _get_new_nlh_env(6)
        env.reset()
        action = env.get_random_action()
        assert (isinstance(action, tuple) and len(action) == 2)

    def test_get_and_set_env(self):
        for n in range(TestPokerEnv.MIN_P, TestPokerEnv.MAX_P + 1):
            env = _get_new_nlh_env(n, 100, 1000, True)
            env.reset()
            for _ in range(TestPokerEnv.ITERATIONS):

                # save state after some actions
                repeat_cause_terminal = True
                while repeat_cause_terminal:
                    o_obs, reward, terminal, info = env.reset()
                    i = 0
                    while not terminal and i < np.random.randint(low=0, high=n * 6):
                        o_obs, reward, terminal, info = env.step(env.get_random_action())
                        i += 1
                        repeat_cause_terminal = terminal

                saved_state = env.state_dict()

                # manipulate env to test if old state is fully restored
                i = 0
                while not terminal and i < np.random.randint(low=0, high=n * 4):
                    obs, reward, terminal, info = env.step(action=env.get_random_action())
                    i += 1

                env.load_state_dict(saved_state)
                obs = env.get_current_obs(False)
                np.array_equal(obs, o_obs)

                env.reset()

    def test_get_hole_cards_of_player(self):
        for n in range(TestPokerEnv.MIN_P, TestPokerEnv.MAX_P + 1):
            env = _get_new_nlh_env(n, 100, 1000)
            env.reset()
            for p in range(env.N_SEATS):
                h = env.get_hole_cards_of_player(p_id=p)
                _h = env.seats[p].hand
                assert np.array_equal(h, _h)

    def test_get_fraction_of_pot_raise(self):
        for n_plyrs in range(TestPokerEnv.MIN_P, TestPokerEnv.MAX_P + 1):
            n = np.random.randint(low=0, high=12)
            for f in [0.1, 0.3, 0.5, 0.7, 1.0, 1.4, 2.2]:
                for _ in range(TestPokerEnv.ITERATIONS):
                    env = _get_new_nlh_env(n_plyrs, 100, 1000)
                    env.reset()

                    for _ in range(n_plyrs):
                        env.step((1, -1))

                    last_raiser_id = env.current_player.seat_id
                    for _ in range(n):  # step a couple times to avoid only testing edge cases
                        # env.step((2, np.random.randint(low=0, high=100)))
                        last_raiser_id = env.current_player.seat_id
                        env.step((2, np.random.randint(low=0, high=180)))

                    next_raiser_id = env.current_player.seat_id

                    suggested_raise = env.get_fraction_of_pot_raise(fraction=f,
                                                                    player_that_bets=env.seats[next_raiser_id])

                    s = sum(env.side_pots) + env.main_pot + sum([p.current_bet for p in env.seats])
                    s += env.seats[last_raiser_id].current_bet - env.seats[next_raiser_id].current_bet

                    last_raiser_tocall_if_next_bets = suggested_raise - env.seats[last_raiser_id].current_bet
                    should_be = s * f

                    assert should_be - 1 <= last_raiser_tocall_if_next_bets <= should_be + 1

    def test_get_frac_from_chip_amt(self):
        for i in range(self.ITERATIONS):
            for n_plyrs in range(TestPokerEnv.MIN_P, TestPokerEnv.MAX_P + 1):
                n = np.random.randint(low=0, high=2)
                for amt in [10, 233, 412, 5001]:
                    env = _get_new_nlh_env(n_plyrs, 300, 1000)
                    env.reset()

                    for _ in range(n_plyrs):
                        env.step((1, -1))

                    for _ in range(n):  # step a couple times to avoid only testing edge cases
                        env.step((2, np.random.randint(low=0, high=255)))
                        if np.random.random() > 0.5:
                            env.step((1, -1))

                    next_raiser_id = env.current_player.seat_id

                    frac = env.get_frac_from_chip_amt(amt=amt, player_that_bets=env.seats[next_raiser_id])
                    suggested_raise = env.get_fraction_of_pot_raise(fraction=frac,
                                                                    player_that_bets=env.seats[next_raiser_id])

                    assert np.allclose(amt, suggested_raise, atol=1.1)

    def test_get_filtered_action_but_change_nothing(self):
        env = _get_new_nlh_env(7, 100, 1000)
        env.reset()

        for _ in range(TestPokerEnv.ITERATIONS):

            orig_action = (np.random.randint(low=0, high=2), np.random.randint(low=0, high=env.MAX_CHIPS))
            _orig_action = copy.deepcopy(orig_action)

            a = env._get_fixed_action(action=orig_action)

            obs, rew, terminal, _ = env.step(action=tuple(a))

            if terminal:
                env.reset()

            assert orig_action == _orig_action


def _get_new_nlh_env(n_seats, min_stack=100, max_stack=1000, random_stacks=False):
    r_m = 0
    if random_stacks:
        r_m = min_stack - max_stack
    args = NoLimitHoldem.ARGS_CLS(n_seats=n_seats,
                                  stack_randomization_range=(r_m, 0),
                                  starting_stack_sizes_list=[1000] * n_seats)
    return NoLimitHoldem(env_args=args, is_evaluating=True, lut_holder=NoLimitHoldem.get_lut_holder())


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Eric Steinberger


import unittest
from unittest import TestCase

import numpy as np

from PokerRL.game._.tree.PublicTree import PublicTree
from PokerRL.game.games import StandardLeduc, DiscretizedNLLeduc
from PokerRL.game.wrappers import HistoryEnvBuilder


class TestGameTree(TestCase):

    def test_building(self):
        _get_leduc_tree()
        _get_nl_leduc_tree()

    def test_vs_env_obs(self):
        for game in ["limit", "nl"]:
            if game == "limit":
                env, env_args = _get_new_leduc_env()
                dummy_env, env_args = _get_new_leduc_env()
                tree = _get_leduc_tree(env_args=env_args)
            else:
                env, env_args = _get_new_nl_leduc_env()
                dummy_env, env_args = _get_new_nl_leduc_env()
                tree = _get_nl_leduc_tree(env_args=env_args)

            lut_holder = StandardLeduc.get_lut_holder()

            env.reset()
            dummy_env.reset()
            node = tree.root

            # RAISE .. stays preflop
            legal = env.get_legal_actions()
            a = 2
            assert a in legal

            o, r, d, i = env.step(a)
            node = node.children[legal.index(a)]
            dummy_env.load_state_dict(node.env_state)
            tree_o = dummy_env.get_current_obs(is_terminal=False)

            env.print_obs(o)
            env.print_obs(tree_o)
            assert np.array_equal(o, tree_o)

            # CALL .. goes flop
            legal = env.get_legal_actions()
            a = 1
            assert a in legal

            o, r, d, i = env.step(a)
            node = node.children[legal.index(1)]
            card_that_came_in_env = lut_holder.get_1d_card(env.board[0])
            node = node.children[card_that_came_in_env]
            dummy_env.load_state_dict(node.env_state)
            tree_o = dummy_env.get_current_obs(is_terminal=False)

            assert np.array_equal(o, tree_o)

            # RAISE .. stays flop
            legal = env.get_legal_actions()
            a = legal[-1]
            assert a in legal

            o, r, d, i = env.step(a)
            node = node.children[legal.index(a)]
            dummy_env.load_state_dict(node.env_state)
            tree_o = dummy_env.get_current_obs(is_terminal=False)

            assert np.array_equal(o, tree_o)


def _get_leduc_tree(env_args=None):
    if env_args is None:
        env_args = StandardLeduc.ARGS_CLS(n_seats=2,
                                          )

    env_bldr = HistoryEnvBuilder(env_cls=StandardLeduc, env_args=env_args)

    _tree = PublicTree(
        env_bldr=env_bldr,
        stack_size=env_args.starting_stack_sizes_list,
        stop_at_street=None
    )

    _tree.build_tree()

    for p in range(env_bldr.N_SEATS):
        _tree.fill_uniform_random()
    _tree.compute_ev()

    _tree.export_to_file()
    print("Tree with stack size", _tree.stack_size, "has", _tree.n_nodes, "nodes out of which", _tree.n_nonterm,
          "are non-terminal.")
    print(np.mean(_tree.root.exploitability) * env_bldr.env_cls.EV_NORMALIZER)

    return _tree


def _get_nl_leduc_tree(env_args=None):
    if env_args is None:
        env_args = DiscretizedNLLeduc.ARGS_CLS(n_seats=2,
                                               starting_stack_sizes_list=[1000, 1000],
                                               bet_sizes_list_as_frac_of_pot=[1.0]
                                               )

    env_bldr = HistoryEnvBuilder(env_cls=DiscretizedNLLeduc, env_args=env_args)

    _tree = PublicTree(
        env_bldr=env_bldr,
        stack_size=env_args.starting_stack_sizes_list,
        stop_at_street=None,
    )

    _tree.build_tree()

    for p in range(env_bldr.N_SEATS):
        _tree.fill_uniform_random()
    _tree.compute_ev()

    _tree.export_to_file()
    print("Tree with stack size", _tree.stack_size, "has", _tree.n_nodes, "nodes out of which", _tree.n_nonterm,
          "are non-terminal.")
    print(np.mean(_tree.root.exploitability) * env_bldr.env_cls.EV_NORMALIZER)

    return _tree


def _get_new_leduc_env(env_args=None):
    if env_args is None:
        env_args = StandardLeduc.ARGS_CLS(n_seats=2,
                                          starting_stack_sizes_list=[150, 150],
                                          )
    return StandardLeduc(env_args=env_args, is_evaluating=True, lut_holder=StandardLeduc.get_lut_holder()), env_args


def _get_new_nl_leduc_env(env_args=None):
    if env_args is None:
        env_args = DiscretizedNLLeduc.ARGS_CLS(n_seats=2,
                                               bet_sizes_list_as_frac_of_pot=[1.0, 1000.0]
                                               )

    return DiscretizedNLLeduc(env_args=env_args, is_evaluating=True,
                              lut_holder=DiscretizedNLLeduc.get_lut_holder()), env_args


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2019 Eric Steinberger


import torch

from PokerRL.rl import rl_util
from PokerRL.rl.base_cls.workers.WorkerBase import WorkerBase


class ParameterServerBase(WorkerBase):
    """
    You do NOT have to use a ParameterServer for your algorithm, but if you do, you can subclass this Base class to
    make use of pre-written functions. You can, however, also create a PS architecture without inheriting from this
    class.
    """

    def __init__(self, t_prof, chief_handle):
        super().__init__(t_prof=t_prof)

        self._env_bldr = rl_util.get_env_builder(t_prof=t_prof)
        self._chief_handle = chief_handle

        self._device = torch.device(t_prof.device_parameter_server)

    def _apply_grads(self, list_of_grads, net, optimizer, grad_norm_clip=None):
        optimizer.zero_grad()
        n_grad_sources = len(list_of_grads)

        # _______________________________________ pull the grads from workers __________________________________________
        # if the code is run distributed, only Object IDs pointing to specific workers arrive, so we have to pull them
        # we batch the .get() calls for stability. Here we create batches
        grad_batches = []
        c = 0
        while n_grad_sources > c:
            s = min(n_grad_sources, c + self._t_prof.max_n_las_sync_simultaneously)
            grad_batches.append(list_of_grads[c:s])
            if type(grad_batches[-1]) is not list:
                grad_batches[-1] = [grad_batches[-1]]
            c = s

        # final grads
        list_of_grads = []
        for batch in grad_batches:
            for g in self._ray.get(batch):
                if g is not None:
                    list_of_grads.append(self._ray.grads_to_torch(g, device=self._device))

        # ________________________________________________ apply grads _________________________________________________
        n_grad_sources = len(list_of_grads)  # have to recompute because the ones that were None are removed now
        if n_grad_sources > 0:
            # reshape and average
            grads = {}
            for name, _ in net.named_parameters():
                grads[name] = []
                for i in range(n_grad_sources):
                    grads[name].append(list_of_grads[i][name])
                grads[name] = torch.mean(torch.stack(grads[name], dim=0), dim=0)

            for name, param in net.named_parameters():
                param.grad = grads[name]

            if grad_norm_clip is not None:
                torch.nn.utils.clip_grad_norm_(parameters=net.parameters(), max_norm=grad_norm_clip)

            optimizer.step()



# Copyright (c) 2019 Eric Steinberger


import numpy as np


class Poker:
    """
    Global constants
    """

    # All common rounds of a poker game. When creating subclasses with different rules, you can remove, but not add
    # rounds from the game by simply overriding ALL_ROUNDS_LIST.
    PREFLOP = 0
    FLOP = 1
    TURN = 2
    RIVER = 3

    INT2STRING_ROUND = {
        PREFLOP: "preflop",
        FLOP: "flop",
        TURN: "turn",
        RIVER: "river",
    }

    STRING2INT_ROUND = {
        "preflop": PREFLOP,
        "flop": FLOP,
        "turn": TURN,
        "river": RIVER
    }

    # Possible actions in a poker game.
    FOLD = 0
    CHECK_CALL = 1
    BET_RAISE = 2

    # The public board has this value wherever a card has not been dealt. 1D and 2D refers to the two different
    # representations of cards: as an int or as a tuple of (rank, suit) respectively.
    CARD_NOT_DEALT_TOKEN_1D = -127
    CARD_NOT_DEALT_TOKEN_2D = np.array([CARD_NOT_DEALT_TOKEN_1D,
                                        CARD_NOT_DEALT_TOKEN_1D])

    # Metrics to measure winnings and exploitability.
    MeasureAnte = "MA_per_G"  # Milli Antes per game
    MeasureBB = "MBB_per_G"  # Milli Big Blinds per game



# Copyright (c) 2019 Eric Steinberger


import numpy as np

from PokerRL.game.Poker import Poker
from PokerRL.game.PokerRange import PokerRange

"""
classes in this file are utilities that poker environments based on PokerEnv can inherit from. They override PokerEnv's
class variables and thereby set it to a certain rule set.
"""


class LeducRules:
    """
    General rules of Leduc Hold'em poker games
    """

    N_HOLE_CARDS = 1
    N_RANKS = 3
    N_SUITS = 2
    N_CARDS_IN_DECK = N_RANKS * N_SUITS
    RANGE_SIZE = PokerRange.get_range_size(n_hole_cards=N_HOLE_CARDS, n_cards_in_deck=N_CARDS_IN_DECK)

    BTN_IS_FIRST_POSTFLOP = True

    N_FLOP_CARDS = 1
    N_TURN_CARDS = 0
    N_RIVER_CARDS = 0
    N_TOTAL_BOARD_CARDS = N_FLOP_CARDS + N_TURN_CARDS + N_RIVER_CARDS
    ALL_ROUNDS_LIST = [Poker.PREFLOP, Poker.FLOP]

    SUITS_MATTER = False

    ROUND_BEFORE = {
        Poker.PREFLOP: Poker.PREFLOP,
        Poker.FLOP: Poker.PREFLOP
    }
    ROUND_AFTER = {
        Poker.PREFLOP: Poker.FLOP,
        Poker.FLOP: None
    }

    RANK_DICT = {i: str(i + 2) for i in range(N_RANKS)}
    SUIT_DICT = {k: ["a", "b", "c", "d", "e", "f", "g"][k] for k in range(N_SUITS)} \
        if N_SUITS < 8 \
        else {i: str(i) for i in range(N_SUITS)}

    STRING = "LEDUC_RULES"

    def __init__(self):
        pass

    def get_hand_rank_all_hands_on_given_boards(self, boards_1d, lut_holder):
        """
        for general docs refer to PokerEnv
        """
        hand_ranks = np.full(shape=(boards_1d.shape[0], LeducRules.RANGE_SIZE), fill_value=-1, dtype=np.int32)
        for board_idx in range(boards_1d.shape[0]):
            for range_idx in range(LeducRules.RANGE_SIZE):
                hand_ranks[board_idx, range_idx] = self.get_hand_rank(
                    hand_2d=lut_holder.get_2d_hole_cards_from_range_idx(range_idx=range_idx),
                    board_2d=lut_holder.get_2d_cards(cards_1d=boards_1d[board_idx]))

        return hand_ranks

    def get_hand_rank(self, hand_2d, board_2d):
        """
        for docs refer to PokerEnv
        """
        if board_2d[0, 0] == hand_2d[0, 0]:
            return 100 + hand_2d[0, 0]
        else:
            return hand_2d[0, 0]

    @classmethod
    def get_lut_holder(cls):
        from PokerRL.game._.look_up_table import LutHolderLeduc

        return LutHolderLeduc(cls)


class BigLeducRules:
    N_HOLE_CARDS = 1
    N_RANKS = 8
    N_SUITS = 3
    N_CARDS_IN_DECK = N_RANKS * N_SUITS
    RANGE_SIZE = PokerRange.get_range_size(n_hole_cards=N_HOLE_CARDS, n_cards_in_deck=N_CARDS_IN_DECK)

    BTN_IS_FIRST_POSTFLOP = True

    N_FLOP_CARDS = 1
    N_TURN_CARDS = 0
    N_RIVER_CARDS = 0
    N_TOTAL_BOARD_CARDS = N_FLOP_CARDS + N_TURN_CARDS + N_RIVER_CARDS
    ALL_ROUNDS_LIST = [Poker.PREFLOP, Poker.FLOP]

    SUITS_MATTER = False

    ROUND_BEFORE = {
        Poker.PREFLOP: Poker.PREFLOP,
        Poker.FLOP: Poker.PREFLOP
    }
    ROUND_AFTER = {
        Poker.PREFLOP: Poker.FLOP,
        Poker.FLOP: None
    }

    RANK_DICT = {i: str(i + 2) for i in range(N_RANKS)}
    SUIT_DICT = {k: ["a", "b", "c", "d", "e", "f", "g"][k] for k in range(N_SUITS)} \
        if N_SUITS < 8 \
        else {i: "_" + str(i) for i in range(N_SUITS)}

    STRING = "BIG_LEDUC_RULES"

    def __init__(self):
        pass

    def get_hand_rank_all_hands_on_given_boards(self, boards_1d, lut_holder):
        """
        for general docs refer to PokerEnv
        """
        hand_ranks = np.full(shape=(boards_1d.shape[0], LeducRules.RANGE_SIZE), fill_value=-1, dtype=np.int32)
        for board_idx in range(boards_1d.shape[0]):
            for range_idx in range(LeducRules.RANGE_SIZE):
                hand_ranks[board_idx, range_idx] = self.get_hand_rank(
                    hand_2d=lut_holder.get_2d_hole_cards_from_range_idx(range_idx=range_idx),
                    board_2d=lut_holder.get_2d_cards(cards_1d=boards_1d[board_idx]))

        return hand_ranks

    def get_hand_rank(self, hand_2d, board_2d):
        """
        for docs refer to PokerEnv
        """
        if board_2d[0, 0] == hand_2d[0, 0]:
            return 100 + hand_2d[0, 0]
        else:
            return hand_2d[0, 0]

    @classmethod
    def get_lut_holder(cls):
        from PokerRL.game._.look_up_table import LutHolderLeduc

        return LutHolderLeduc(cls)


class HoldemRules:
    """
    General rules of Texas Hold'em poker games
    """
    N_HOLE_CARDS = 2
    N_RANKS = 13
    N_SUITS = 4
    N_CARDS_IN_DECK = N_RANKS * N_SUITS
    RANGE_SIZE = PokerRange.get_range_size(n_hole_cards=N_HOLE_CARDS, n_cards_in_deck=N_CARDS_IN_DECK)

    BTN_IS_FIRST_POSTFLOP = False

    N_FLOP_CARDS = 3
    N_TURN_CARDS = 1
    N_RIVER_CARDS = 1
    N_TOTAL_BOARD_CARDS = N_FLOP_CARDS + N_TURN_CARDS + N_RIVER_CARDS
    ALL_ROUNDS_LIST = [Poker.PREFLOP, Poker.FLOP, Poker.TURN, Poker.RIVER]

    SUITS_MATTER = True

    ROUND_BEFORE = {
        Poker.PREFLOP: Poker.PREFLOP,
        Poker.FLOP: Poker.PREFLOP,
        Poker.TURN: Poker.FLOP,
        Poker.RIVER: Poker.TURN
    }
    ROUND_AFTER = {
        Poker.PREFLOP: Poker.FLOP,
        Poker.FLOP: Poker.TURN,
        Poker.TURN: Poker.RIVER,
        Poker.RIVER: None
    }

    RANK_DICT = {
        Poker.CARD_NOT_DEALT_TOKEN_1D: "",
        0: "2",
        1: "3",
        2: "4",
        3: "5",
        4: "6",
        5: "7",
        6: "8",
        7: "9",
        8: "T",
        9: "J",
        10: "Q",
        11: "K",
        12: "A"
    }
    SUIT_DICT = {
        Poker.CARD_NOT_DEALT_TOKEN_1D: "",
        0: "h",
        1: "d",
        2: "s",
        3: "c"
    }

    STRING = "HOLDEM_RULES"

    def __init__(self):
        from PokerRL.game._.cpp_wrappers.CppHandeval import CppHandeval

        self._clib = CppHandeval()

    def get_hand_rank_all_hands_on_given_boards(self, boards_1d, lut_holder):
        """
        for docs refer to PokerEnv
        """
        return self._clib.get_hand_rank_all_hands_on_given_boards_52_holdem(boards_1d=boards_1d, lut_holder=lut_holder)

    def get_hand_rank(self, hand_2d, board_2d):
        """
        for docs refer to PokerEnv
        """
        return self._clib.get_hand_rank_52_holdem(hand_2d=hand_2d, board_2d=board_2d)

    @classmethod
    def get_lut_holder(cls):
        from PokerRL.game._.look_up_table import LutHolderHoldem

        return LutHolderHoldem(cls)


class FlopHoldemRules:
    """
    General rules of Texas Hold'em poker games
    """
    N_HOLE_CARDS = 2
    N_RANKS = 13
    N_SUITS = 4
    N_CARDS_IN_DECK = N_RANKS * N_SUITS
    RANGE_SIZE = PokerRange.get_range_size(n_hole_cards=N_HOLE_CARDS, n_cards_in_deck=N_CARDS_IN_DECK)

    BTN_IS_FIRST_POSTFLOP = False

    N_FLOP_CARDS = 5
    N_TURN_CARDS = 0
    N_RIVER_CARDS = 0
    N_TOTAL_BOARD_CARDS = N_FLOP_CARDS
    ALL_ROUNDS_LIST = [Poker.PREFLOP, Poker.FLOP]

    SUITS_MATTER = True

    ROUND_BEFORE = {
        Poker.PREFLOP: Poker.PREFLOP,
        Poker.FLOP: Poker.PREFLOP,
        Poker.TURN: None,
        Poker.RIVER: None,
    }
    ROUND_AFTER = {
        Poker.PREFLOP: Poker.FLOP,
        Poker.FLOP: None,
        Poker.TURN: None,
        Poker.RIVER: None,
    }

    RANK_DICT = {
        Poker.CARD_NOT_DEALT_TOKEN_1D: "",
        0: "2",
        1: "3",
        2: "4",
        3: "5",
        4: "6",
        5: "7",
        6: "8",
        7: "9",
        8: "T",
        9: "J",
        10: "Q",
        11: "K",
        12: "A"
    }
    SUIT_DICT = {
        Poker.CARD_NOT_DEALT_TOKEN_1D: "",
        0: "h",
        1: "d",
        2: "s",
        3: "c"
    }

    STRING = "FLOP_HOLDEM_RULES"

    def __init__(self):
        from PokerRL.game._.cpp_wrappers.CppHandeval import CppHandeval

        self._clib = CppHandeval()

    def get_hand_rank_all_hands_on_given_boards(self, boards_1d, lut_holder):
        """
        for docs refer to PokerEnv
        """
        return self._clib.get_hand_rank_all_hands_on_given_boards_52_holdem(boards_1d=boards_1d, lut_holder=lut_holder)

    def get_hand_rank(self, hand_2d, board_2d):
        """
        for docs refer to PokerEnv
        """
        return self._clib.get_hand_rank_52_holdem(hand_2d=hand_2d, board_2d=board_2d)

    @classmethod
    def get_lut_holder(cls):
        from PokerRL.game._.look_up_table import LutHolderHoldem

        return LutHolderHoldem(cls)



# Copyright (c) 2019 Eric Steinberger


"""
Wraps the local LBR master in a ray actor to be placed on any (one) machine in the cluster. This worker will then
manage distributed LBRWorkers to compute the local best response approximation faster in parallel.
"""
import ray

from PokerRL.eval.lbr.LocalLBRMaster import LocalLBRMaster as _LocalEvalLBRMaster


@ray.remote(num_cpus=1)
class DistLBRMaster(_LocalEvalLBRMaster):

    def __init__(self, t_prof, chief_handle):
        _LocalEvalLBRMaster.__init__(self, t_prof=t_prof, chief_handle=chief_handle)



# Copyright (c) 2019 Eric Steinberger


import shutil
from os.path import join as ospj

from PokerRL._.CrayonWrapper import CrayonWrapper
from PokerRL.rl.base_cls.workers.WorkerBase import WorkerBase
from PokerRL.util import file_util


class DriverBase(WorkerBase):
    """
    The Driver runs the HighLevelAlgo and creates all workers. If running distributed, the Driver is the worker node
    that launches the computation and distributes tasks between workers.
    """

    def __init__(self, t_prof, eval_methods, chief_cls, eval_agent_cls, n_iterations=None, iteration_to_import=None,
                 name_to_import=None):
        """
        Args:
            t_prof (TrainingProfile)
            eval_methods (dict):                dict of {evaluator1_name: frequency, ...} Currently supported evaluators
                                                are "br", "rlbr", and "lbr"
            chief_cls (ChiefBase subclass):     class, not instance
            n_iterations (int)                  number of iterations to run. If None, runs forever
            iteration_to_import (int):               step/iteration to import
            name_to_import (str):               name of the run to import
        """
        super().__init__(t_prof=t_prof)

        if self._t_prof.CLUSTER:
            self._ray.init_cluster(redis_address=t_prof.redis_head_adr)
        else:
            self._ray.init_local()

        file_util.do_pickle(obj=t_prof, file_name=t_prof.name, path=t_prof.path_trainingprofiles)
        self.n_iterations = n_iterations

        self._step_to_import = iteration_to_import
        self._name_to_import = name_to_import

        if self._t_prof.DISTRIBUTED:
            from PokerRL.eval.lbr.DistLBRMaster import DistLBRMaster as LBRMaster
            from PokerRL.eval.rl_br.DistRLBRMaster import DistRLBRMaster as RLBRMaster
            from PokerRL.eval.rl_br.workers.ps.Dist_RLBR_ParameterServer import \
                Dist_RLBR_ParameterServer as RLBRParameterServer
            from PokerRL.eval.rl_br.workers.la.Dist_RLBR_LearnerActor import Dist_RLBR_LearnerActor as RLBRLearnerActor
            from PokerRL.eval.lbr.DistLBRWorker import DistLBRWorker as LBRWorker
            from PokerRL.eval.br.DistBRMaster import DistBRMaster as BRMaster
            from PokerRL.eval.head_to_head.DistHead2HeadMaster import DistHead2HeadMaster as Head2HeadMaster

        else:
            from PokerRL.eval.lbr.LocalLBRMaster import LocalLBRMaster as LBRMaster
            from PokerRL.eval.rl_br.LocalRLBRMaster import LocalRLBRMaster as RLBRMaster
            from PokerRL.eval.rl_br.workers.ps.Local_RLBR_ParameterServer import \
                Local_RLBR_ParameterServer as RLBRParameterServer
            from PokerRL.eval.rl_br.workers.la.Local_RLBR_LearnerActor import \
                Local_RLBR_LearnerActor as RLBRLearnerActor
            from PokerRL.eval.lbr.LocalLBRWorker import LocalLBRWorker as LBRWorker
            from PokerRL.eval.br.LocalBRMaster import LocalBRMaster as BRMaster
            from PokerRL.eval.head_to_head.LocalHead2HeadMaster import LocalHead2HeadMaster as Head2HeadMaster

        # safety measure to avoid overwriting logs when reloading
        if name_to_import is not None and iteration_to_import is not None and name_to_import == t_prof.name:
            t_prof.name += "_"

        print("Creating Chief...")
        self.chief_handle = self._ray.create_worker(chief_cls,
                                                    t_prof)

        self.eval_masters = {}
        if "br" in list(eval_methods.keys()):
            print("Creating BR Evaluator...")
            self.eval_masters["br"] = (self._ray.create_worker(BRMaster,
                                                               t_prof,
                                                               self.chief_handle,
                                                               eval_agent_cls),
                                       eval_methods["br"]  # freq
                                       )

        if "h2h" in list(eval_methods.keys()):
            print("Creating Head-to-Head Mode Evaluator...")
            self.eval_masters["h2h"] = (self._ray.create_worker(Head2HeadMaster,
                                                                t_prof,
                                                                self.chief_handle,
                                                                eval_agent_cls),
                                        eval_methods["h2h"]  # freq
                                        )

        if "lbr" in list(eval_methods.keys()):
            print("Creating LBR Evaluator...")
            self._lbr_workers = [
                self._ray.create_worker(LBRWorker,
                                        t_prof,
                                        self.chief_handle,
                                        eval_agent_cls)
                for _ in range(self._t_prof.module_args["lbr"].n_workers)
            ]

            self.eval_masters["lbr"] = (self._ray.create_worker(LBRMaster,
                                                                t_prof,
                                                                self.chief_handle),
                                        eval_methods["lbr"]  # freq
                                        )
            self._ray.wait([
                self._ray.remote(self.eval_masters["lbr"][0].set_worker_handles,
                                 *self._lbr_workers)
            ])

        if "rlbr" in list(eval_methods.keys()):
            print("Creating RL-BR Evaluator...")
            self._rlbr_ps = self._ray.create_worker(RLBRParameterServer,
                                                    t_prof,
                                                    self.chief_handle,
                                                    )
            self._rlbr_las = [
                self._ray.create_worker(RLBRLearnerActor,
                                        t_prof,
                                        self.chief_handle,
                                        eval_agent_cls
                                        )
                for _ in range(self._t_prof.module_args["rlbr"].n_las)
            ]

            self.eval_masters["rlbr"] = (self._ray.create_worker(RLBRMaster,
                                                                 t_prof,
                                                                 self.chief_handle,
                                                                 eval_agent_cls),
                                         eval_methods["rlbr"]  # freq
                                         )

            self._ray.wait([
                self._ray.remote(self.eval_masters["rlbr"][0].set_learner_actors,
                                 *self._rlbr_las),
            ])
            self._ray.wait([
                self._ray.remote(self.eval_masters["rlbr"][0].set_param_server,
                                 self._rlbr_ps),
            ])

        self.crayon = CrayonWrapper(name=t_prof.name, chief_handle=self.chief_handle,
                                    path_log_storage=self._t_prof.path_log_storage,
                                    crayon_server_address=t_prof.local_crayon_server_docker_address,
                                    runs_distributed=t_prof.DISTRIBUTED,
                                    runs_cluster=t_prof.CLUSTER,
                                    )

    def _maybe_load_checkpoint_init(self):
        if self._step_to_import is None:
            self._cfr_iter = 0
        else:
            assert self._step_to_import is not None
            print("Loading checkpoint ", self._step_to_import)
            self._cfr_iter = self._step_to_import + 1
            self.load_checkpoint(step=self._step_to_import, name_to_load=self._name_to_import)
            print("Loaded from iter: ", self._step_to_import)

    def _delete_past_checkpoints(self, steps_not_to_delete):
        _dir = ospj(self._t_prof.path_checkpoint, self._t_prof.name)
        all_dir_names = file_util.get_all_dirs_in_dir(_dir)
        dir_names_to_delete = [e for e in all_dir_names if e not in [str(s) for s in steps_not_to_delete]]

        for step in dir_names_to_delete:
            shutil.rmtree(ospj(_dir, step))

    def _delete_past_log_files(self, steps_not_to_delete):
        _dir = ospj(self._t_prof.path_log_storage, self._t_prof.name)
        all_dir_names = file_util.get_all_dirs_in_dir(_dir)
        dir_names_to_delete = [e for e in all_dir_names if e not in [str(s) for s in steps_not_to_delete]]

        for step in dir_names_to_delete:
            shutil.rmtree(ospj(_dir, step))

    def evaluate(self):
        """
        puts whole network on wait while the parameters are synced, but evaluates while training the next iteration(s)
        """
        evaluators_to_run = []
        for kind in list(self.eval_masters.keys()):
            ev = self.eval_masters[kind][0]
            freq = self.eval_masters[kind][1]
            if self._cfr_iter % freq == 0:
                evaluators_to_run.append(ev)
                print("Evaluating vs.", kind.upper())
                self._ray.wait([
                    self._ray.remote(ev.update_weights)
                ])

        for ev in evaluators_to_run:
            self._ray.remote(ev.evaluate,
                             self._cfr_iter)

    def save_logs(self):
        """
        Pulls all new logs from the LogBuffer and adds them to crayon (Tensorboard wrapper). It also saves them to disk.
        """
        self.crayon.update_from_log_buffer()
        self.crayon.export_all(iter_nr=self._cfr_iter)

        # Delete past version
        s = [self._cfr_iter]
        if self._cfr_iter > self._t_prof.log_export_freq + 1:
            s.append(self._cfr_iter - self._t_prof.log_export_freq)

        self._delete_past_log_files(steps_not_to_delete=s)

    def periodically_export_eval_agent(self):
        if self._cfr_iter % self._t_prof.eval_agent_export_freq == 0:
            self.export_eval_agent()

    def periodically_checkpoint(self):
        if self._cfr_iter % self._t_prof.checkpoint_freq == 0:
            print("Saving Checkpoint")
            self.checkpoint(curr_step=self._cfr_iter)

    def export_eval_agent(self):
        print("Exporting agent")
        self._ray.wait([
            self._ray.remote(self.chief_handle.export_agent,
                             self._cfr_iter)
        ])

    # ____________________________________________________ Override ____________________________________________________
    def run(self):
        """
        Calling this function should start up your algorithm.
        """
        raise NotImplementedError

    def checkpoint(self, curr_step):
        """
        Calling this function should trigger "".checkpoint(curr_step=curr_step)"" on all workers in the network.
        """
        raise NotImplementedError

    def load_checkpoint(self, name_to_load, step):
        """
        Calling this function should trigger all workers to load ""name_to_load"" at iteration ""step"" by calling
        worker.load_checkpoint(name_to_load=name_to_load, step=step)
        """
        raise NotImplementedError



# Copyright (c) 2019 Eric Steinberger


import numpy as np


class AgentTournament:

    def __init__(self, env_cls, env_args, eval_agent_1, eval_agent_2):
        self._eval_agents = [eval_agent_1, eval_agent_2]

        self._env_cls = env_cls
        self._env_args = env_args
        self._lut_holder = self._env_cls.get_lut_holder()

        assert env_args.n_seats == 2

    def run(self, n_games_per_seat):
        REFERENCE_AGENT = 0
        _env = self._env_cls(env_args=self._env_args, is_evaluating=True, lut_holder=self._lut_holder)

        winnings = np.empty(shape=(n_games_per_seat * _env.N_SEATS), dtype=np.float32)

        for seat_p0 in range(_env.N_SEATS):
            seat_p1 = 1 - seat_p0

            for _hand_nr in range(n_games_per_seat):
                # """""""""""""""""
                # Reset
                # """""""""""""""""
                _, r_for_all, done, info = _env.reset()
                for e in self._eval_agents:
                    e.reset(deck_state_dict=_env.cards_state_dict())

                # """""""""""""""""
                # Play Episode
                # """""""""""""""""
                while not done:
                    p_id_acting = _env.current_player.seat_id

                    if p_id_acting == seat_p0:
                        action_int, _ = self._eval_agents[REFERENCE_AGENT].get_action(step_env=True, need_probs=False)
                        self._eval_agents[1 - REFERENCE_AGENT].notify_of_action(p_id_acted=p_id_acting,
                                                                                action_he_did=action_int)
                    elif p_id_acting == seat_p1:
                        action_int, _ = self._eval_agents[1 - REFERENCE_AGENT].get_action(step_env=True,
                                                                                          need_probs=False)
                        self._eval_agents[REFERENCE_AGENT].notify_of_action(p_id_acted=p_id_acting,
                                                                            action_he_did=action_int)
                    else:
                        raise ValueError("Only HU supported!")

                    _, r_for_all, done, info = _env.step(action_int)

                # """""""""""""""""
                # Add Rews
                # """""""""""""""""
                winnings[_hand_nr + (seat_p0 * n_games_per_seat)] = r_for_all[seat_p0] \
                                                                    * _env.REWARD_SCALAR \
                                                                    * _env.EV_NORMALIZER
        mean = np.mean(winnings).item()
        std = np.std(winnings).item()

        _d = 1.96 * std / np.sqrt(n_games_per_seat * _env.N_SEATS)
        lower_conf95 = mean - _d
        upper_conf95 = mean + _d

        print()
        print("Played", n_games_per_seat * 2, "hands of poker.")
        print("Player ", self._eval_agents[REFERENCE_AGENT].get_mode() + ":", mean, "+/-", _d)
        print("Player ", self._eval_agents[1 - REFERENCE_AGENT].get_mode() + ":", (-mean), "+/-", _d)

        return float(mean), float(upper_conf95), float(lower_conf95)



# Copyright (c) 2019 Eric Steinberger


import numpy as np


class PokerPlayer:
    """
    Holds information about the player state in a PokerEnv and gives an interface to modify player variables.
    """

    def __init__(self, seat_id, poker_env, is_evaluating, starting_stack, stack_randomization_range):
        assert isinstance(seat_id, int)
        assert isinstance(starting_stack, int)
        assert isinstance(stack_randomization_range, tuple)
        assert len(stack_randomization_range) == 2

        self.seat_id = seat_id
        self.poker_env = poker_env
        self._base_starting_stack = starting_stack
        self.stack_randomization_range = stack_randomization_range

        self.IS_EVALUATING = is_evaluating

        # player vars
        self.hand = None  # np.ndarray
        self.hand_rank = None
        self.current_bet = None
        self.stack = None
        self.starting_stack_this_episode = None

        # flags for table management
        self.is_allin = None
        self.folded_this_episode = None
        self.has_acted_this_round = None

        # each player has his own side-pot rank.
        # rank0 gets only the main_pot. rank1 gets main-pot and self.side_pots[1], ...
        self.side_pot_rank = None

        # sets all the above
        self.reset()

    def reset(self):

        if self.IS_EVALUATING or (self.stack_randomization_range[0] == 0 and self.stack_randomization_range[1] == 0):
            self.starting_stack_this_episode = self._base_starting_stack
        else:
            self.starting_stack_this_episode = \
                max(
                    self.poker_env.BIG_BLIND,
                    np.random.randint(
                        low=self._base_starting_stack - np.abs(self.stack_randomization_range[0]),
                        high=self._base_starting_stack + self.stack_randomization_range[1] + 1)
                )

        # episode info
        self.stack = self.starting_stack_this_episode
        self.hand = []  # [[rank, suit], ...]
        self.current_bet = 0

        # flags for table management
        self.is_allin = False
        self.folded_this_episode = False
        self.has_acted_this_round = False
        self.side_pot_rank = -1

    def bet_raise(self, total_bet_amount):
        """
        assumes bet is 100% legal and bet_size <= stack size

        Args:
            total_bet_amount (int)
        """

        self.has_acted_this_round = True
        self.stack -= (total_bet_amount - self.current_bet)
        self.current_bet = total_bet_amount

        if self.stack == 0:
            self.is_allin = True

    def check_call(self, total_to_call):
        """
        Assumes having total_to_call as self.current_bet is 100% legal and just does it.

        Args:
            total_to_call (int)
        """
        self.has_acted_this_round = True
        delta = int(total_to_call - self.current_bet)
        self.stack -= delta
        self.current_bet = int(total_to_call)  # equal to += delta but is faster

        if self.stack == 0:
            self.is_allin = True

    def fold(self):
        self.has_acted_this_round = True
        self.folded_this_episode = True

    def award(self, amount):
        self.stack += amount

    def refund_from_bet(self, amount):
        self.stack += amount
        self.current_bet -= amount

    def player_state(self):
        return (self.seat_id,
                self.stack,
                self.folded_this_episode,
                self.is_allin)
